//************************************************************************************************* 
//																								  *
// File: CommandLineProcess.cs																	  * 
// ---------------------------																	  * 
//																								  *
// ---------------------------------------------------------------------------------------------- * 
//																								  * 
// Description: CommandLineProcess class. Code and XML comments.						          * 
// -----------																			          * 
//																								  *
// Namespace:	twComps																		      * 
// ---------																		        	  * 
//																								  *
// Classes:																						  * 
// -------																						  * 
//																								  *
// - CommandLineProcess:	Provides a programmatical interface for executing a command line.	  * 
//							Textual output and return value are accessible by various			  *	
//							properties and events.												  *	
//																								  * 
// ---------------------------------------------------------------------------------------------- * 
//																								  * 
// Author:			Thomas Weller																  * 
// Created:			2004/04/25																	  * 
// Last change:		2004/11/30																	  * 
// Last change by:	Thomas Weller																  * 
// 																		                          * 
//************************************************************************************************* 

namespace twComps
{
	using System;
	using System.Diagnostics;
	using System.Threading;
	using System.Globalization;
	using System.Text;

	/// <summary>
	/// Encapsulates a system process that can execute a command line and provides access
	/// to the resulting textual output. 
	/// </summary>
	/// <remarks>
	/// The <b>CommandLineProcess</b> class is maybe best described as a 'windowless console'. It is 
	/// a component class designed for calling an external command line from within a Windows Forms
	/// application (or any other application that wants to avoid opening a console window) and
	/// receiving its output. This works for stand-alone command line applications as well as for
	/// commands adressing the Windows command line processor ('DOS commands'). 
	/// <para/>
	/// Whether or not the Windows command line processor shall be used to execute a specific command
	/// is controlled by the objects <see cref="UseComSpec"/> property. The internal handling of 
	/// the process strongly differs when calling/not calling the Windows command line processor.
	/// (The details are described in the topics for the related members, see 
	/// <see href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
	/// Start</see>, <see cref="Abort"/>, <see cref="ReadProcessOutput"/>.) However, for the caller it
	/// is totally transparent.
	/// <para/>
	/// For execution, <b>CommandLineProcess</b> makes heavy use of an encapsulated
	/// <see cref="System.Diagnostics.Process">System.Diagnostics.Process</see> member.
	/// Many of <b>CommandLineProcess'</b> properties and operations are simply mirorring
	/// corresponding members of this encapsulated system process.
	/// <para/>
	/// The <b>CommandLineProcess</b> class avoids throwing any exceptions. Instead of throwing
	/// exceptions, the class signals error conditions by returning <c>false</c> (methods) or
	/// some predefined error values (properties). Exceptions, that may be caused by other code
	/// (e.g. from the <b>System.Process</b> instance), are caught internally and will not be
	/// noticeable to a caller of <b>CommandLineProcess</b> code.
	/// <para/>
	/// Besides utilizing the execution of a command line, the second purpose of the 
	/// <b>CommandLineProcess</b> class is to provide a handy mechanism for dealing with the resulting output.
	/// Every line of text that is outputted by the encapsulated system process during execution is internally
	/// caught by a <b>CommandLineProcess</b> instance and further broadcasted using an <see cref="OutputLine"/> 
	/// event. Any 'DOS'-related text elements will be suppressed during this process, if necessary. (I.e.
	/// What <b>OutputLine</b> will deliver is always exactly the same as one would read from the console
	/// window when having entered the command manually.)<br/>
	/// To monitor the system process' output (there are two output streams: 
	/// <see cref="System.Diagnostics.Process.StandardOutput">StandardOutput</see> and
	/// <see cref="System.Diagnostics.Process.StandardError">StandardError</see>), <b>CommandLineProcess</b>
	/// starts a separate thread together with the command line process, using an encapsulated
	/// <see cref="System.Threading.Thread"> System.Threading.Thread</see> object.
	/// <para/>
	/// The <b>CommandLineProcess</b> class exposes various different events, that can be evaluated by other
	/// processes to keep track of <b>CommandLineProcess'</b> state:
	/// <list type="bullet">
	/// <item>
	/// <see cref="Started"/> signals the successful start of the command line process.
	/// </item>
	/// <item>
	/// <see cref="Aborted"/> signals that a running command line process was killed.
	/// </item>
	/// <item>
	/// <see cref="Exited"/> signals that the command line process has exited.
	/// </item>
	/// <item>
	/// <see cref="Closed"/> signals that all system resources allocated for the process component were freed.
	/// </item>
	/// <item>
	/// <see cref="OutputLine"/> broadcasts one line of text read from one of the system process' 
	/// output streams.
	/// </item>
	/// <item>
	/// <see cref="CmdLineChanged"/> signals that the object's command line text, or its destinated
	/// usage, has changed.
	/// </item>
	/// </list>
	/// For any of these events, <b>CommandLineProcess</b> declares a corresponding delegate type for
	/// handling the event. To associate an event with an event handler, add an instance of the delegate 
	/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
	/// For more information about handling events with delegates, see 
	/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
	/// Events and Delegates</see>.
	/// <note>
	/// Any code that processes an <see cref="OutputLine"/> event must be safe for multi-threading!
	/// </note>
	/// <para/>
	/// Another handy feature of the <b>CommandLineProcess</b> class is the ability to encode process output
	/// using any available code page set by using the object's <see cref="OutputCodepage"/> property. A 
	/// <see cref="System.Text.Encoding">System.Text.Encoding</see> object is used to do that. This is 
	/// especially useful if you have to take care of correctly displaying language-specific characters 
	/// like e.g. german umlauts.
	/// <para/>
	/// <note>
	/// Which code page and/or output stream a command line application is using under which circumstances,
	/// is not really predictable. Never rely on the fact that an application will write its error 
	/// messages to the <b>StandardError</b> stream and its normal output to the <b>StandardOutput</b> 
	/// stream! This may be true in many cases, but in many others it is not. You have to check this 
	/// out for every single application ...
	/// </note>
	/// </remarks>
	/// <example>
	/// The following example demonstrates the usage of the <b>CommandLineProcess</b> class in
	/// a Windows Forms application.<br/>
	/// Note that this example will work with the <b>UseComSpec</b> property set to <c>true</c>
	/// as well as <c>false</c>, since <b>ping</b> is a 'DOS command', but it is at the same time
	/// an exe-file that can be called directly.
	/// <br/>
	/// <code>
	/// 
	/// class CommandLineProcessDemo_form
	/// {
	///		CommandLineProcess m_cmd;
	///		
	///		... 
	///		
	///		RichTextBox richEditStdOutput;
	///		RichTextBox richEditStdError;
	///		
	///		... 
	///		
	///		[STAThread]
	///		static void Main() 
	///		{
	///			Application.Run(new CommandLineProcessDemo_form());
	///		}
	///		
	///		public CommandLineProcessDemo_form() 
	///		{
	///			// form initialization stuff ...
	///			
	///			...
	///			
	///			m_cmd = new CommandLineProcess();
	///			
	///			// Connect to CommandLineProcess' events
	///			// -------------------------------------
	///			m_cmd.CmdLineChanged += new CommandLineProcess.CmdLineChangedEventHandler(OnCmdLineChanged);
	///			m_cmd.Aborted        += new CommandLineProcess.AbortedEventHandler(OnAborted);
	///			m_cmd.Exited         += new CommandLineProcess.ExitedEventHandler(OnExited);
	///			m_cmd.Started        += new CommandLineProcess.StartedEventHandler(OnStarted);
	///			m_cmd.OutputLine     += new CommandLineProcess.OutputLineEventHandler(OnOutputLine);
	///			m_cmd.Closed         += new CommandLineProcess.ClosedEventHandler(OnClosed);
	///			
	///			// do something...
	///			PingToYourself();
	///		}
	///		
	///		... 
	///		
	///		// call the ping utility
	///		private void PingToYourself()
	///		{
	///			m_cmd.Command = "ping";
	///			m_cmd.Arguments = "localhost";
	///			m_cmd.UseComSpec = true;		// this is optionally here
	///			if (!m_cmd.Start()) 
	///				MessageBox.Show("CommandLineProcess.Start() failed!");
	///		}
	///		
	///		... 
	///		
	///		// Protect main window from being closed while CommandLineProcess is executing
	///		private void CommandLineProcessDemo_form_Closing(object sender, System.ComponentModel.CancelEventArgs e)
	///		{
	///			if (m_cmd.IsRunning) 
	///			{
	///				e.Cancel = true; // Cancel closing
	///				MessageBox.Show("Closing canceled, CommandLineProcess is executing. Wait for exit or abort."); 
	///			}
	///		}
	///		
	///		... 
	///		
	///		// Event handlers to receive CommandLineProcess' events
	///		// ----------------------------------------------------
	///		
	///		// process outputs one line of text
	///		private void OnOutputLine(object sender, CommandLineProcess.OutputLineEventArgs e)
	///		{
	///			RichTextBox rtb;
	///
	///			// which output stream ?
	///			if (e.OutputType == CommandLineProcess.OutputType.StdOut) 
	///				rtb = this.richEditStdOutput;
	///			else if (e.OutputType == CommandLineProcess.OutputType.StdErr)
	///				rtb = this.richEditStdError;
	///			else return;
	///
	///			// ignore empty lines
	///			if ((e.Text == null) || (e.Text.Length == 0)) return;
	///			
	///			lock(this)	// protect richtextbox variable from being accessed by different threads simultaneously
	///			{
	///				rtb.AppendText(e.Text + "\n"); // write to richedit control
	///			}
	///			
	///			if (e.OutputType == CommandLineProcess.OutputType.StdOut) 
	///				Debug.WriteLine("StandardOutput: " + e.Text); 
	///			else if (e.OutputType == CommandLineProcess.OutputType.StdErr)
	///				Debug.WriteLine("StandardError: " + e.Text); 
	///		}
	///
	///		// CommandLineProcess has exited
	///		private void OnExited(object sender, CommandLineProcess.ExitedEventArgs e)
	///		{
	///			StringBuilder strBld = new StringBuilder("Exited event received, Time");
	///			
	///			if (e.ExitTime != CommandLineProcess.UndefinedTimeValue)
	///			{
	///				strBld.Append(": " + e.ExitTime.ToString("hh:mm:ss.ff") + ", Code");
	///			}
	///			else
	///			{
	///				strBld.Append(" not defined, Code");
	///			}
	///			
	///			if (e.ExitCode != CommandLineProcess.UndefinedCodeValue)
	///			{
	///				strBld.Append(": " + e.ExitCode.ToString() + ".");
	///			}
	///			else
	///			{
	///				strBld.Append(" not defined.");
	///			}
	///			
	///			Debug.WriteLine(strBld.ToString()); 
	///			
	///			m_cmd.Command = null;	// clear command line
	///		}
	///
	///		// Process was closed (i.e. all related system resources were freed)
	///		private void OnClosed(object sender, CommandLineProcess.ClosedEventArgs e)
	///		{
	///			if (e.CloseTime == CommandLineProcess.UndefinedTimeValue)
	///			{
	///				Debug.WriteLine("Closed event received, Time not defined."); 
	///			}
	///			else
	///			{
	///				Debug.WriteLine("Closed event received, Time: '" + 
	///								 e.CloseTime.ToString("hh:mm:ss.ff") + "'."); 
	///			}
	///		}
	///
	///		// Command line has changed
	///		private void OnCmdLineChanged(object sender, CommandLineProcess.CmdLineChangedEventArgs e)
	///		{
	///			if (e.Text != null)
	///			{
	///				Debug.WriteLine("CmdLineChanged event received, Text: '" + e.Text + "'."); 
	///			}
	///			else
	///			{
	///				Debug.WriteLine("CmdLineChanged event received, Text is null."); 
	///			}
	///		}
	///
	///		// CommandLineProcess has started
	///		private void OnStarted(object sender, CommandLineProcess.StartedEventArgs e)
	///		{
	///			if (e.StartTime == CommandLineProcess.UndefinedTimeValue)
	///			{
	///				Debug.WriteLine("Started event received, Time not defined."); 
	///			}
	///			else
	///			{
	///				Debug.WriteLine("Started event received, Time: '" + 
	///								 e.StartTime.ToString("hh:mm:ss.ff") + "'."); 
	///			}
	///		}
	///
	///		// CommandLineProcess has aborted
	///		private void OnAborted(object sender, CommandLineProcess.AbortedEventArgs e)
	///		{
	///			if (e.AbortTime == CommandLineProcess.UndefinedTimeValue)
	///			{
	///				Debug.WriteLine("Aborted event received, Time not defined."); 
	///			}
	///			else
	///			{
	///				Debug.WriteLine("Aborted event received, Time: '" + 
	///								 e.AbortTime.ToString("hh:mm:ss.ff") + "'."); 
	///			}
	///		}
	/// }
	/// 
	/// </code>
	/// <br/>
	/// This code will produce a debug output similar to the following
	/// (on a german machine named 'centrino', using codepage 850):
	/// <br/><br/><c>
	/// CmdLineChanged event received, Text: 'ping'. UseComSpec: No.<br/>
	/// CmdLineChanged event received, Text: 'ping localhost'. UseComSpec: No.<br/>
	/// CmdLineChanged event received, Text: 'ping localhost'. UseComSpec: Yes.<br/>
	/// Started event received, Time: '01:58:19.70'.<br/>
	/// <![CDATA[StandardOutput: Ping centrino [127.0.0.1] mit 32 Bytes Daten:]]><br/>
	/// <![CDATA[StandardOutput: Antwort von 127.0.0.1: Bytes=32 Zeit<1ms TTL=64]]><br/>
	/// <![CDATA[StandardOutput: Antwort von 127.0.0.1: Bytes=32 Zeit<1ms TTL=64]]><br/>
	/// <![CDATA[StandardOutput: Antwort von 127.0.0.1: Bytes=32 Zeit<1ms TTL=64]]><br/>
	/// <![CDATA[StandardOutput: Antwort von 127.0.0.1: Bytes=32 Zeit<1ms TTL=64]]><br/>
	/// StandardOutput: Ping-Statistik fr 127.0.0.1:<br/>
	/// StandardOutput:     Pakete: Gesendet = 4, Empfangen = 4, Verloren = 0 (0% Verlust),<br/>
	/// StandardOutput: Ca. Zeitangaben in Millisek.:<br/>
	/// StandardOutput:     Minimum = 0ms, Maximum = 0ms, Mittelwert = 0ms<br/>
	/// Exited event received, Time: 01:58:22.95, Code: 0.<br/>
	/// CmdLineChanged event received, Text is null.<br/>
	/// </c>
	/// </example>
	/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
	/// <seealso cref="System.Diagnostics.Process.StandardOutput">System.Diagnostics.Process.StandardOutput</seealso>
	/// <seealso cref="System.Diagnostics.Process.StandardError">System.Diagnostics.Process.StandardError</seealso>
	/// <seealso cref="System.Threading.Thread">System.Threading.Thread</seealso>
	/// <seealso cref="System.Text.Encoding">System.Text.Encoding</seealso>
	public class CommandLineProcess
	{
	
	#region Event Handler Delegates and Event Argument Classes
	
		/// <summary>
		/// Represents the method that will handle the <see cref="CmdLineChanged"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// A <see cref="CmdLineChangedEventArgs">CmdLineChangedEventArgs</see> object that holds the new 
		/// command line string in its <see cref="CmdLineChangedEventArgs.Text">Text</see> member. The 
		/// <see cref="CmdLineChangedEventArgs.UseComSpec">UseComSpec</see> field indicates whether this
		/// command shall be treated as a stand-alone command line or as a 'DOS command'.
		/// </param>
		/// <remarks>
		/// A <see cref="CmdLineChanged"/> event is fired by a <see cref="CommandLineProcess"/> instance
		/// whenever its internal command line (or its usage) was changed by using its <see cref="Command"/>,
		/// <see cref="Arguments"/> or <see cref="UseComSpec"/> properties.	The
		/// <see cref="CmdLineChangedEventArgs.Text">Text</see> member of the	
		/// <b>CmdLineChangedEventArgs</b> object will contain the new command line string.
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>CmdLineChangedEventHandler</b> delegate declaration.<br/>
		/// When you create a <b>CmdLineChangedEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="CmdLineChangedEventArgs">CmdLineChangedEventArgs</seealso>
		/// <seealso cref="CommandLineProcess.CmdLineChangedEventArgs.Text">CmdLineChangedEventArgs.Text</seealso>
		/// <seealso cref="CommandLineProcess.CmdLineChangedEventArgs.Text">CmdLineChangedEventArgs.Text</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.CmdLineChanged">CommandLineProcess.CmdLineChanged</seealso>
		/// <seealso cref="CommandLineProcess.CommandLine">CommandLineProcess.CommandLine</seealso>
		/// <seealso cref="CommandLineProcess.Command">CommandLineProcess.Command</seealso>
		/// <seealso cref="CommandLineProcess.Arguments">CommandLineProcess.Arguments</seealso>
		/// <seealso cref="CommandLineProcess.UseComSpec">CommandLineProcess.UseComSpec</seealso>
		public delegate void CmdLineChangedEventHandler(object sender, CmdLineChangedEventArgs e);
		/// <summary>
		/// Represents the method that will handle the <see cref="Started"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// A <see cref="StartedEventArgs">StartedEventArgs</see> object that holds the starting time
		/// for the process in its <see cref="StartedEventArgs.StartTime"/> member.
		/// </param>
		/// <remarks>
		/// A <b>Started</b> event is fired by a <see cref="CommandLineProcess"/> instance whenever one of its
		/// <see href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
		/// Start</see> methods was successfully called. Refer to these methods for details.
		/// <para/>
		/// The <b>StartTime</b> member of the <b>StartedEventArgs</b> object will contain the exact start time 
		/// of the process.<br/>
		/// Normally, <b>StartedEventArgs.StartTime</b> will be the same value that can be retrieved via the
		/// <see cref="CommandLineProcess.StartTime">CommandLineProcess.StartTime</see> property after starting.
		/// However, in some cases, <b>StartTime</b> might be set to <see cref="UndefinedTimeValue">
		/// CommandLineProcess.UndefinedTimeValue</see>. This occurs if the start time was not returned by the
		/// underlying <see cref="System.Diagnostics.Process">Process</see> object, but the preceding call to its
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// Start</see> method returned true. 
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>StartedEventHandler</b> delegate declaration.<br/>
		/// When you create a <b>StartedEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="StartedEventArgs">StartedEventArgs</seealso>
		/// <seealso cref="StartedEventArgs.StartTime">StartedEventArgs.StartTime</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
		/// CommandLineProcess.Start Overload List</seealso>
		/// <seealso cref="CommandLineProcess.Started">CommandLineProcess.Started</seealso>
		/// <seealso cref="CommandLineProcess.StartTime">CommandLineProcess.StartTime</seealso>
		/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public delegate void StartedEventHandler(object sender, StartedEventArgs e);
		/// <summary>
		/// Represents the method that will handle the <see cref="Exited"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// An <see cref="ExitedEventArgs">ExitedEventArgs</see> object that holds the exit time
		/// for the process in its <see cref="ExitedEventArgs.ExitTime"/> member.
		/// </param>
		/// <remarks>
		/// An <b>Exited</b> event is fired by a <see cref="CommandLineProcess"/> instance, 
		/// whenever it receives an <see cref="System.Diagnostics.Process.Exited">Exited</see> event
		/// from its underlying <see cref="System.Diagnostics.Process">Process</see> object.
		/// <para/>
		/// The <b>ExitTime</b> member of the <b>ExitedEventArgs</b> object will contain
		/// the exact exit time.
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>ExitedEventHandler</b> delegate declaration.<br/>
		/// When you create an <b>ExitedEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="ExitedEventArgs">ExitedEventArgs</seealso>
		/// <seealso cref="CommandLineProcess.ExitedEventArgs.ExitTime">ExitedEventArgs.ExitTime</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Exited">CommandLineProcess.Exited</seealso>
		/// <seealso cref="CommandLineProcess.ExitTime">CommandLineProcess.ExitTime</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitTime">System.Diagnostics.Process.ExitTime</seealso>
		public delegate void ExitedEventHandler(object sender, ExitedEventArgs e);
		/// <summary>
		/// Represents the method that will handle the <see cref="Aborted"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// An <see cref="AbortedEventArgs">AbortedEventArgs</see> object that holds the abort time
		/// for the process in its <see cref="AbortedEventArgs.AbortTime"/> member.
		/// </param>
		/// <remarks>
		/// An <b>Aborted</b> event is fired by a <see cref="CommandLineProcess"/> instance
		/// whenever <see cref="CommandLineProcess.Abort"/> was successfully called. Refer to this 
		/// method for details.
		/// <para/>
		/// The <b>AbortTime</b> member of the <b>AbortedEventArgs</b> object will contain the exact 
		/// abort time of the process. This is the time, when a call to the 
		/// <see cref="System.Diagnostics.Process.Kill">Kill</see> method of the underlying 
		/// <see cref="System.Diagnostics.Process">System.Diagnostics.Process</see>
		/// object was successful, i.e. did not throw an exception.
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>AbortedEventHandler</b> delegate declaration.<br/>
		/// When you create an <b>AbortedEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="AbortedEventArgs">AbortedEventArgs</seealso>
		/// <seealso cref="AbortedEventArgs.AbortTime">AbortedEventArgs.AbortTime</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Aborted">CommandLineProcess.Aborted</seealso>
		/// <seealso cref="CommandLineProcess.Abort">CommandLineProcess.Abort</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</seealso>
		public delegate void AbortedEventHandler(object sender, AbortedEventArgs e);
		/// <summary>
		/// Represents the method that will handle the <see cref="Closed"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// A <see cref="ClosedEventArgs"/> object that holds the time when the process
		/// was closed in its <see cref="ClosedEventArgs.CloseTime"/> member.
		/// </param>
		/// <remarks>
		/// A <b>Closed</b> event is fired by a <see cref="CommandLineProcess"/> instance
		/// whenever <see cref="CommandLineProcess.Close"/> was successfully called. Refer to this 
		/// method for details.
		/// <para/>
		/// The <b>CloseTime</b> member of the <b>ClosedEventArgs</b> object will contain the exact closing
		/// time of the process.<br/>
		/// This is the time, when a call to the <see cref="System.Diagnostics.Process.Close">
		/// Close</see> method of the underlying <see cref="System.Diagnostics.Process">Process
		/// </see> object was successful, i.e. did not throw an exception.
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>ClosedEventHandler</b> delegate declaration.<br/>
		/// When you create a <b>ClosedEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="ClosedEventArgs"/>
		/// <seealso cref="ClosedEventArgs.CloseTime">ClosedEventArgs.CloseTime</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Closed">CommandLineProcess.Closed</seealso>
		/// <seealso cref="CommandLineProcess.Close">CommandLineProcess.Close</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.Close">System.Diagnostics.Process.Close</seealso>
		public delegate void ClosedEventHandler(object sender, ClosedEventArgs e);
		/// <summary>
		/// Represents the method that will handle the <see cref="OutputLine"/> event.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// An <see cref="OutputLineEventArgs">OutputLineEventArgs</see> object that holds one line of
		/// textual output in its <see cref="OutputLineEventArgs.Text"/> field together with a flag indicating
		/// whether this line was written to <b>StandardError</b> or <b>StandardOutput</b>. The 
		/// <see cref="OutputLineEventArgs.OutputType">OutputType</see> member field contains information
		/// about which stream was written to.
		/// </param>
		/// <remarks>
		/// An <see cref="OutputLine"/> event is fired by a <see cref="CommandLineProcess"/> instance
		/// for every line of text that the underlying <see cref="System.Diagnostics.Process">Process</see>
		/// object outputs to either <b>StandardError</b> or <b>StandardOutput</b>.<br/>
		/// This output is intercepted by the <b>CommandLineProcess</b> object and re-encoded using
		/// the requested codepage. See <see cref="CommandLineProcess.OutputCodepage"/> for details.
		/// <note>
		/// Whether a command line tool writes to <b>StandardError</b> or <b>StandardOutput</b>, cannot 
		/// be predicted, but should be checked for every single case. Usage of these two output streams 
		/// is quite chaotic.
		/// </note>
		/// <para/>
		/// The declaration of your event handler must have the same parameters as the 
		/// <b>OutputLineEventHandler</b> delegate declaration.<br/>
		/// When you create a <b>OutputLineEventHandler</b> delegate, you identify the method that will 
		/// handle the event. To associate the event with your event handler, add an instance of the delegate 
		/// to the event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.
		/// </remarks>
		/// <seealso cref="OutputLineEventArgs">OutputLineEventArgs</seealso>
		/// <seealso cref="OutputLineEventArgs.Text">OutputLineEventArgs.Text</seealso>
		/// <seealso cref="OutputLineEventArgs.OutputType">OutputLineEventArgs.OutputType</seealso>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.OutputLine">CommandLineProcess.OutputLine</seealso>
		/// <seealso cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</seealso>
		/// <seealso cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.StandardError">System.Diagnostics.Process.StandardError</seealso>
		/// <seealso cref="System.Diagnostics.Process.StandardOutput">System.Diagnostics.Process.StandardOutput</seealso>
		public delegate void OutputLineEventHandler(object sender, OutputLineEventArgs e);

		/// <summary>
		/// Provides data for the <see cref="OutputLine"/> event.
		/// </summary>
		/// <remarks>
		/// An <b>OutputLineEventArgs</b> instance holds one line of textual output in its
		/// <see cref="OutputLineEventArgs.Text">Text</see> field together with a flag - 
		/// <see cref="OutputLineEventArgs.OutputType">OutputType</see> - indicating whether
		/// this line was written to the process' <b>StandardError</b> or <b>StandardOutput</b> stream.
		/// <para/>
		/// An <see cref="OutputLine"/> event is fired by a <b>CommandLineProcess</b> instance for every
		/// line of text that the underlying <see cref="System.Diagnostics.Process">Process</see> 
		/// object outputs to either <b>StandardError</b> or <b>StandardOutput</b>.<br/>
		/// This string may be null or an empty string.<br/>
		/// If the <see cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</see> 
		/// property is set to a value that represents a valid code page number, <b>Text</b> will be encoded 
		/// using that code page.
		/// <note>
		/// Whether a command line tool writes to <b>StandardError</b> or <b>StandardOutput</b>, cannot be
		/// predicted, but should be checked for every single case. 
		/// Usage of these two output streams is quite chaotic.
		/// </note>
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.OutputLine">CommandLineProcess.OutputLine</seealso>
		/// <seealso cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</seealso>
		/// <seealso cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</seealso>
		/// <seealso cref="System.Diagnostics.Process.StandardOutput">System.Diagnostics.Process.StandardOutput</seealso>
		/// <seealso cref="System.Diagnostics.Process.StandardError">System.Diagnostics.Process.StandardError</seealso>
		public class OutputLineEventArgs: EventArgs 
		{
			/// <summary>
			/// Indicates whether the line was written to the process' <b>StandardError</b>
			/// or <b>StandardOutput</b> stream.
			/// </summary>
			/// <remarks>
			/// See <see cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</see> enumeration.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</seealso>
			/// <seealso cref="System.Diagnostics.Process.StandardOutput">System.Diagnostics.Process.StandardOutput</seealso>
			/// <seealso cref="System.Diagnostics.Process.StandardError">System.Diagnostics.Process.StandardError</seealso>
			public readonly OutputType OutputType;
			/// <summary>
			/// Contains one line of output.
			/// </summary>
			/// <remarks>
			/// The value of <b>Text</b> might be <c>null</c> or an empty string, if this was produced by
			/// the underlying <see cref="System.Diagnostics.Process">Process</see> object to
			/// print an extra line.
			/// <para/>
			/// If the <see cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</see> 
			/// property is set to a value that represents a valid code page number, <b>Text</b> will be 
			/// encoded using that code page.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</seealso>
			public readonly string Text;

			/// <overloads>
			/// Constructs an <b>OutputLineEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an <b>OutputLineEventArgs</b> instance and sets its <b>Text</b> and
			/// <b>OutputType</b> members.
			/// </summary>
			/// <param name="type">
			/// Indicates whether the line was written to StandardOutput or StandardError.
			/// </param>
			/// <param name="text">
			/// Line of text that was written to output stream, eventually re-encoded (using
			/// <see cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</see>).
			/// </param>
			/// <remarks>
			/// Sets the <b>Text</b> and <b>OutputType</b> members.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.OutputCodepage">CommandLineProcess.OutputCodepage</seealso>
			/// <seealso cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</seealso>
			public OutputLineEventArgs(OutputType type, string text)			
			{ 
				OutputType = type;
				Text = text;
			}
			/// <summary>
			/// Constructs an <b>OutputLineEventArgs</b> instance and sets its <b>OutputType</b> member.
			/// </summary>
			/// <param name="type">
			/// Indicates whether the line was written to StandardOutput or StandardError.
			/// </param>
			/// <remarks>
			/// The <b>Text</b> member is set to <c>null</c>, indicating an empty line.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.OutputType">CommandLineProcess.OutputType</seealso>
			public OutputLineEventArgs(OutputType type)			
			{ 
				OutputType = type;
				Text = null;
			}
		}

		/// <summary>
		/// Provides data for the <see cref="CmdLineChanged"/> event.
		/// </summary>
		/// <remarks>
		/// An <b>CmdLineChangedEventArgs</b> instance holds the command line as it is currently
		/// stored in the <see cref="CommandLineProcess"/> instance in its
		/// <see cref="CmdLineChangedEventArgs.Text">Text</see> field. The 
		/// <see cref="CmdLineChangedEventArgs.UseComSpec">UseComSpec</see> field indicates whether this
		/// command is treated as a stand-alone command line or as a 'DOS command'.
		/// <para/>
		/// A <see cref="CmdLineChanged"/> event is fired by a <see cref="CommandLineProcess"/> instance
		/// whenever its internal command line (or its usage) was changed by using its <see cref="Command"/>,
		/// <see cref="Arguments"/> or <see cref="UseComSpec"/> properties.		
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.CmdLineChanged">CommandLineProcess.CmdLineChanged</seealso>
		/// <seealso cref="CommandLine">CommandLineProcess.CommandLine</seealso>
		/// <seealso cref="Command">CommandLineProcess.Command</seealso>
		/// <seealso cref="Arguments">CommandLineProcess.Arguments</seealso>
		/// <seealso cref="UseComSpec">CommandLineProcess.UseComSpec</seealso>
		public class CmdLineChangedEventArgs: EventArgs 
		{
			/// <summary>
			/// The current command line of the <see cref="CommandLineProcess"/> instance.
			/// </summary>
			/// <remarks>
			/// A value of <c>null</c> denotes an empty command line.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.CommandLine">CommandLineProcess.CommandLine</seealso>
			public readonly string Text;

			/// <summary>
			/// Flag indicating whether the Windows command line processor should be used.
			/// </summary>
			/// <remarks>
			/// A value of <c>true</c> means that the command line is treated as a 'DOS command', i.e.
			/// the command line as a whole is passed as argument to <c>cmd.exe</c>.<br/>
			/// See <see cref="CommandLineProcess.UseComSpec">CommandLineProcess.UseComSpec</see> 
			/// for details.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.UseComSpec">CommandLineProcess.UseComSpec</seealso>
			public readonly bool UseComSpec;

			/// <overloads>
			/// Constructs a <b>CmdLineChangedEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an empty <b>CmdLineChangedEventArgs</b>.
			/// </summary>
			/// <remarks>
			/// The <b>Text</b> member is set to <c>null</c>. <b>UseComSpec</b> is set to <c>false</c>.
			/// </remarks>
			public CmdLineChangedEventArgs() { Text = null; UseComSpec = false; }
			/// <summary>
			/// Constructs an <b>CmdLineChangedEventArgs</b> instance and sets its <b>Text</b> member.
			/// </summary>
			/// <param name="text">
			/// Value of the <b>Text</b> member.
			/// </param>
			/// <remarks>
			/// The <b>UseComSpec</b> member is set to <c>false</c>.
			/// </remarks>
			public CmdLineChangedEventArgs(string text)	{ Text = text; UseComSpec = false; }
			
			/// <summary>
			/// Constructs an <b>CmdLineChangedEventArgs</b> instance and sets its <b>Text</b> 
			/// and <b>UseComSpec</b> members.
			/// </summary>
			/// <param name="text">
			/// Value of the <b>Text</b> member.
			/// </param>
			/// <param name="useComSpec">
			/// Value of the <b>UseComSpec</b> member.
			/// </param>
			public CmdLineChangedEventArgs(string text, bool useComSpec)	
													{ Text = text; UseComSpec = useComSpec; }
		}

		/// <summary>
		/// Provides data for the <see cref="Started"/> event.
		/// </summary>
		/// <remarks>
		/// The <b>StartTime</b> member contains the exact start time of the process.<br/>
		/// Normally, <b>StartTime</b> will be the same value that can be retrieved via the
		/// <see cref="CommandLineProcess.StartTime">CommandLineProcess.StartTime</see> 
		/// property after starting.<br/>
		/// However, in some case, <b>StartTime</b> might be set to <see cref="UndefinedTimeValue">
		/// CommandLineProcess.UndefinedTimeValue</see>. This occurs if the start time was not 
		/// returned by the underlying <see cref="System.Diagnostics.Process">Process</see> object,
		/// but the preceding call to 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// Process.Start</see> returned true. 
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Start">CommandLineProcess.Start</seealso>
		/// <seealso cref="CommandLineProcess.Started">CommandLineProcess.Started</seealso>
		/// <seealso cref="CommandLineProcess.StartTime">CommandLineProcess.StartTime</seealso>
		/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public class StartedEventArgs: EventArgs 
		{
			/// <summary>
			/// The start time of the process as retrieved from the underlying
			/// <see cref="System.Diagnostics.Process">Process</see> object when started.
			/// </summary>
			/// <remarks>
			/// <b>StartTime</b> reflects the value of the <see cref="System.Diagnostics.Process.StartTime"> 
			/// Process.StartTime</see> property immediately after starting the process. 
			/// <para/>
			/// If unknown or invalid, <b>StartTime</b> will have the value
			/// <see cref="UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.StartTime">CommandLineProcess.StartTime</seealso>
			/// <seealso cref="System.Diagnostics.Process.StartTime">System.Diagnostics.Process.StartTime</seealso>
			public readonly DateTime StartTime;

			/// <overloads>
			/// Constructs a <b>StartedEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an empty <b>StartedEventArgs</b>.
			/// </summary>
 			/// <remarks>
			/// The <b>StartTime</b> member is set to <see cref="UndefinedTimeValue">
			/// CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			public StartedEventArgs()
			{
				StartTime = UndefinedTimeValue;
			}
			/// <summary>
			/// Constructs a <b>StartedEventArgs</b> instance and sets its <b>StartTime</b> member.
			/// </summary>
			/// <param name="time">
			/// Value of the <b>StartTime</b> member.
			/// </param>
			public StartedEventArgs(DateTime time)
			{
				StartTime = time;
			}
		}

		/// <summary>
		/// Provides data for the <see cref="Aborted"/> event.
		/// </summary>
		/// <remarks>
		/// The <see cref = "AbortTime"/> member contains the time when <see cref = "CommandLineProcess"/> 
		/// successfully called the <see cref="System.Diagnostics.Process.Kill">Kill</see> method of the 
		/// underlying <see cref="System.Diagnostics.Process">Process</see> object (i.e. <b>Kill</b> did
		/// not raise an exception when called).
		/// <para/>
		/// If abortion was successful, an <see cref="Exited"/> event will be sent by the <b>CommandLineProcess</b>
		/// object in succession. If there was pending text output from the process at the time abortion took
		/// place, the appropriate <see cref="OutputLine"/> event(s) will be sent before the <b>Exited</b> event.
		/// <note>
		/// The value of <b>AbortTime</b> does not necessarily mean that the underlying process was really
		/// (immediately) killed. In this respect, you should rely exclusively on the <b>Exited</b> event.
		/// </note>
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Abort">CommandLineProcess.Abort</seealso>
		/// <seealso cref="CommandLineProcess.Aborted">CommandLineProcess.Aborted</seealso>
		/// <seealso cref="CommandLineProcess.Exited">CommandLineProcess.Exited</seealso>
		/// <seealso cref="CommandLineProcess.OutputLine">CommandLineProcess.OutputLine</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</seealso>
		public class AbortedEventArgs: EventArgs 
		{
			/// <summary>
			/// The time when abortion took place.
			/// </summary>
 			/// <remarks>
			/// If unknown or invalid, <b>AbortTime</b> will have the value
			/// <see cref="UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.Abort">CommandLineProcess.Abort</seealso>
			/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>
			public readonly DateTime AbortTime;

			/// <overloads>
			/// Constructs an <b>AbortedEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an empty <b>AbortedEventArgs</b>.
			/// </summary>
 			/// <remarks>
			/// The <b>AbortTime</b> member is set to <see cref="UndefinedTimeValue">
			/// CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			public AbortedEventArgs()
			{
				AbortTime = UndefinedTimeValue;
			}
			/// <summary>
			/// Constructs an <b>AbortedEventArgs</b> instance and sets its <b>AbortTime</b> member.
			/// </summary>
			/// <param name="time">
			/// Value of the <b>AbortTime</b> member.
			/// </param>
 			/// <remarks>
			/// The <b>AbortTime</b> member is set to <b>time</b>.
			/// </remarks>
			public AbortedEventArgs(DateTime time)
			{
				AbortTime = time;
			}
		}

		/// <summary>
		/// Provides data for the <see cref="Closed"/> event.
		/// </summary>
		/// <remarks>
		/// The <b>CloseTime</b> member contains the time when <see cref="CommandLineProcess"/> successfully
		/// called the <see cref="System.Diagnostics.Process.Close">Close</see> method of the underlying
		/// <see cref="System.Diagnostics.Process">Process</see> object (i.e. <b>Close</b> did not raise 
		/// an exception when called).
		/// <para/>
		/// A call to <b>Close</b> frees all the resources that are associated with a process component. 
		/// Consequently, the various <b>CommandLineProcess</b> properties won't return any more values 
		/// after a call to <b>Close</b>.
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Close">CommandLineProcess.Close</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.Close">System.Diagnostics.Process.Close</seealso>
		public class ClosedEventArgs: EventArgs 
		{
			/// <summary>
			/// The time when the process was closed.
			/// </summary>
 			/// <remarks>
			/// If unknown or invalid, <b>CloseTime</b> will have the value
			/// <see cref="UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.Close">CommandLineProcess.Close</seealso>
			/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>
			public readonly DateTime CloseTime;

			/// <overloads>
			/// Constructs an <b>ClosedEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an empty <b>ClosedEventArgs</b>.
			/// </summary>
 			/// <remarks>
			/// The <b>CloseTime</b> member is set to <see cref="UndefinedTimeValue">
			/// CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			public ClosedEventArgs()
			{
				CloseTime = UndefinedTimeValue;
			}
			/// <summary>
			/// Constructs an <b>ClosedEventArgs</b> instance and sets its <b>CloseTime</b> member.
			/// </summary>
			/// <param name="time">
			/// Value of the <b>CloseTime</b> member.
			/// </param>
 			/// <remarks>
			/// The <b>CloseTime</b> member is set to <b>time</b>.
			/// </remarks>
			public ClosedEventArgs(DateTime time)
			{
				CloseTime = time;
			}
		}

		/// <summary>
		/// Provides data for the <see cref="Exited"/> event.
		/// </summary>
		/// <remarks>
		/// The <b>CommandLineProcess.Exited</b> event mirrors the 
		/// <see cref="System.Diagnostics.Process.Exited">System.Diagnostics.Process.Exited</see> event.
		/// Consequently, the <b>ExitCode</b> and <b>ExitTime</b> members of the <b>ExitedEventArgs</b> class 
		/// contain the values of the corresponding properties retrieved from the underlying 
		/// <see cref="System.Diagnostics.Process">System.Diagnostics.Process</see> component
		/// at the time when the process component sent this event.
		/// </remarks>
		/// <seealso cref="CommandLineProcess">CommandLineProcess class</seealso>
		/// <seealso cref="CommandLineProcess.Exited">CommandLineProcess.Exited</seealso>
		/// <seealso cref="CommandLineProcess.ExitCode">CommandLineProcess.ExitCode</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitCode">System.Diagnostics.Process.ExitCode</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitTime">System.Diagnostics.Process.ExitTime</seealso>
		public class ExitedEventArgs: EventArgs 
		{
			/// <summary>
			/// Exit code of the process.
			/// </summary>
			/// <remarks>
			/// <b>ExitCode</b> contains the value that was retrieved from the
			/// <see cref="System.Diagnostics.Process.ExitCode">ExitCode</see> property of the
			/// underlying <see cref="System.Diagnostics.Process">System.Diagnostics.Process
			/// </see> component at the time when the component sent the <b>Exited</b> event.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.ExitCode">CommandLineProcess.ExitCode</seealso>
			/// <seealso cref="System.Diagnostics.Process.ExitCode">System.Diagnostics.Process.ExitCode</seealso>
			public readonly int ExitCode;
			/// <summary>
			/// Exit time of the process.
			/// </summary>
			/// <remarks>
			/// <b>ExitTime</b> contains the value that was retrieved from the
			/// <see cref="System.Diagnostics.Process.ExitTime">ExitTime</see> property of the 
			/// underlying <see cref="System.Diagnostics.Process">System.Diagnostics.Process
			/// </see> component at the time when the component sent the <b>Exited</b> event.
			/// <para/>
			/// If unknown or invalid, <b>ExitTime</b> will have the value
			/// <see cref="UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</see>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.ExitTime">CommandLineProcess.ExitTime</seealso>
			/// <seealso cref="System.Diagnostics.Process.ExitTime">System.Diagnostics.Process.ExitTime</seealso>
			/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>.
			public readonly DateTime ExitTime;

			/// <overloads>
			/// Constructs an <b>ExitedEventArgs</b> instance in various ways.
			/// </overloads>
			/// <summary>
			/// Constructs an empty <b>ExitedEventArgs</b>.
			/// </summary>
 			/// <remarks>
			/// The <b>ExitTime</b> member is set to <see cref="UndefinedTimeValue">
			/// CommandLineProcess.UndefinedTimeValue</see>, the <b>ExitCode</b> member 
			/// to <see cref="UndefinedCodeValue">CommandLineProcess.UndefinedCodeValue</see>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>.
			/// <seealso cref="CommandLineProcess.UndefinedCodeValue">CommandLineProcess.UndefinedCodeValue</seealso>.
			public ExitedEventArgs()
			{
				ExitCode = UndefinedCodeValue;
				ExitTime = UndefinedTimeValue;
			}
			/// <summary>
			/// Constructs an <b>ExitedEventArgs</b> instance and sets its 
			/// <b>ExitCode</b> and <b>ExitTime</b> members.
			/// </summary>
			/// <param name="code">
			/// Value of the <b>ExitCode</b> member.
			/// </param>
			/// <param name="time">
			/// Value of the <b>ExitTime</b> member.
			/// </param>
 			/// <remarks>
			/// Sets the <b>ExitCode</b> and <b>ExitTime</b> members.
			/// </remarks>
			public ExitedEventArgs(int code, DateTime time)
			{
				ExitCode = code;
				ExitTime = time;
			}
			/// <summary>
			/// Constructs an <b>ExitedEventArgs</b> instance and sets its 
			/// <b>ExitCode</b> member.
			/// </summary>
			/// <param name="code">
			/// Value of the <b>ExitCode</b> member.
			/// </param>
 			/// <remarks>
			/// The <b>ExitTime</b> member is set to <see cref="UndefinedTimeValue"/>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.UndefinedTimeValue">CommandLineProcess.UndefinedTimeValue</seealso>.
			public ExitedEventArgs(int code)
			{
				ExitCode = code;
				ExitTime = UndefinedTimeValue;
			}
			/// <summary>
			/// Constructs an <b>ExitedEventArgs</b> instance and sets its 
			/// <b>ExitTime</b> member.
			/// </summary>
			/// <param name="time">
			/// Value of the <b>ExitTime</b> member.
			/// </param>
 			/// <remarks>
			/// The <b>ExitCode</b> member is set to <see cref="UndefinedCodeValue"/>.
			/// </remarks>
			/// <seealso cref="CommandLineProcess.UndefinedCodeValue">CommandLineProcess.UndefinedCodeValue</seealso>.
			public ExitedEventArgs(DateTime time)
			{
				ExitCode = UndefinedCodeValue;
				ExitTime = time;
			}
		}


	#endregion // Event Handler Delegates and Event Argument Classes 

	#region Construction/Destruction

		/// <overloads>
		/// Constructs a <b>CommandLineProcess</b> instance in various ways.
		/// </overloads>
		/// <summary>
		/// Initializes a <b>CommandLineProcess</b> instance.
		/// </summary>
		/// <remarks>
		/// Creates and initializes a <b>CommandLineProcess</b> instance.
		/// </remarks>
		/// <seealso cref = "Init">Init (protected)</seealso>
		public CommandLineProcess()
		{
			Init(); // object initialization
		}

		/// <summary>
		/// Initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b> property.
		/// </summary>
		/// <param name="command">
		/// Value of the <see cref="Command"/> property.
		/// </param>
		/// <remarks>
		/// Creates and initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b> 
		/// property.
		/// </remarks>
		/// <seealso cref = "Init">Init (protected)</seealso>
		/// <seealso cref="Command"/>
		public CommandLineProcess(string command)
		{
			Init();									// object initialization
			m_proc.StartInfo.FileName = command;	// set Command, but don't cause an event to be raised
		}

		/// <summary>
		/// Initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b>
		/// and <b>Arguments</b> properties.
		/// </summary>
		/// <param name="command">
		/// Value of the <see cref="Command"/> property.
		/// </param>
		/// <param name="arguments">
		/// Value of the <see cref="Arguments"/> property.
		/// </param>
		/// <remarks>
		/// Creates and initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b> 
		/// and <b>Arguments</b> property.
		/// </remarks>
		/// <seealso cref = "Init">Init (protected)</seealso>
		/// <seealso cref="Command"/>
		/// <seealso cref="Arguments"/>
		public CommandLineProcess(string command, string arguments)
		{
			Init();										// object initialization
			m_proc.StartInfo.FileName  = command;		// Set Command and Arguments, but don't 
			m_proc.StartInfo.Arguments = arguments;		//  cause an event to be raised.
		}

		/// <summary>
		/// Initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b>, <b>Arguments</b>, 
		/// and <b>UseComSpec</b> properties.
		/// </summary>
		/// <param name="command">
		/// Value of the <see cref="Command"/> property.
		/// </param>
		/// <param name="arguments">
		/// Value of the <see cref="Arguments"/> property.
		/// </param>
		/// <param name="useComSpec">
		/// </param>
		/// Value of the <see cref="UseComSpec"/> property.
		/// <remarks>
		/// Creates and initializes a <b>CommandLineProcess</b> instance and sets its <b>Command</b>, 
		/// <b>Arguments</b>, and <b>UseComSpec</b> property.
		/// </remarks>
		/// <seealso cref = "Init">Init (protected)</seealso>
		/// <seealso cref="Command"/>
		/// <seealso cref="Arguments"/>
		/// <seealso cref="UseComSpec"/>
		public CommandLineProcess(string command, string arguments, bool useComSpec)
		{
			Init();										// object initialization
			m_proc.StartInfo.FileName  = command;		// Set Command, Arguments, and 
			m_proc.StartInfo.Arguments = arguments;		//  UseComSpec, but don't cause 
			this.m_bUseComSpec = useComSpec;			//  an event to be raised.
		}


	#endregion //  Construction/Destruction

	#region Fields

		/// <summary>
		/// Holds the command portion of the command line.
		/// </summary>
		/// <remarks>
		/// The <b>m_strCommand</b> field is backing the <see cref="Command"/> property.
		/// </remarks>
		/// <seealso cref="Command"/>
		protected string m_strCommand = null;

		/// <summary>
		/// Holds the arguments portion of the command line.
		/// </summary>
		/// <remarks>
		/// The <b>m_strArguments</b> field is backing the <see cref="Arguments"/> property.
		/// </remarks>
		/// <seealso cref="Command"/>
		protected string m_strArguments = null;

		/// <summary>
		/// Holds a value specifying whether to use ComSpec or not.
		/// </summary>
		/// <remarks>
		/// The <b>m_bUseComSpec</b> field is backing the <see cref="UseComSpec"/> property.
		/// </remarks>
		/// <seealso cref="UseComSpec"/>
		protected bool m_bUseComSpec = false;

		/// <summary>
		/// The <see cref="System.Diagnostics.Process">System.Diagnostics.Process</see>
		/// object used for execution. 
		/// </summary>
		/// <remarks>
		/// This is the object actually executing the command line.
		/// </remarks>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process class</seealso>
		protected Process m_proc = new Process();
		/// <summary>
		/// The <see cref="System.Threading.Thread">System.Threading.Thread</see>
		/// object used to receive textual output from the process.
		/// </summary>
		/// <remarks>
		/// The thread object is associated with the <see cref="ReadProcessOutput"/> delegate. Since the
		/// thread's purpose is to intercept process output, it is started/stopped parallel to that process.
		/// See <see cref="StartOutputRecording"/> and <see cref="StopOutputRecording"/>.
		/// </remarks>
		/// <seealso cref="System.Threading.Thread">System.Threading.Thread</seealso>
		/// <seealso cref="ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		/// <seealso cref="StartOutputRecording">StartOutputRecording (protected)</seealso>
		/// <seealso cref="ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		protected Thread m_threadRec = null;

		/// <summary>
		/// The <see cref="System.Text.Encoding">System.Text.Encoding</see> 
		/// object used to re-encode process output.
		/// </summary>
		/// <remarks>
		/// Used to re-encode process output according to the value of the
		/// <see cref="OutputCodepage"/> property.
		/// </remarks>
		/// <seealso cref="System.Text.Encoding">System.Text.Encoding</seealso>
		/// <seealso cref="OutputCodepage"/>
		protected Encoding m_encOutput = null;

		/// <summary>
		/// Temporary buffer used in the process of re-encoding text.
		/// </summary>
		/// <remarks>
		/// This static buffer is used to speed up encoding. It is initialized to 1K size.
		/// </remarks>
		protected byte[] m_buffer = new byte[1024]; // 1K buffer should always be big enough ...

		// Internal flag used to suppress 'DOS-related' parts of text output.
		private bool m_bBlockDosOutput = false;

	#endregion // Fields

	#region Constants and Enumerations

		/// <summary>
		/// Constant <see cref="int"/> value to denote an invalid return code.
		/// </summary>
		/// <remarks>
		/// <b>UndefinedCodeValue</b> is returned in an error case by the <see cref="ExitCode"/> 
		/// property instead of raising an exception.
		/// <para/>
		/// The value of <b>UndefinedCodeValue</b> is <b>int.MinValue</b>, i.e. <b>-2,147,483,648</b>.
		/// </remarks>
		public static readonly int UndefinedCodeValue = int.MinValue;			// -2,147,483,648
		/// <summary>
		/// Constant <see cref="DateTime"/> value to denote an invalid time.
		/// </summary>
		/// <remarks>
		/// <b>UndefinedTimeValue</b> is returned in an error case by the <see cref="StartTime"/> 
		/// and <see cref="ExitTime"/> properties instead of raising an exception.
		/// <para/>
		/// The value of <b>UndefinedTimeValue</b> is <b>DateTime.MinValue</b>, 
		/// i.e. <b>00:00:00.0, January 1, 0001</b>.
		/// </remarks>
		public static readonly DateTime UndefinedTimeValue = DateTime.MinValue;	// 00:00:00.0, January 1, 0001
		/// <summary>
		/// Constant <see cref="IntPtr"/> value to denote an invalid handle value.
		/// </summary>
		/// <remarks>
		/// <b>UndefinedHandleValue</b> is returned in an error case by the <see cref="Handle"/> property 
		/// instead of raising an exception.
		/// <para/>
		/// The value of <b>UndefinedHandleValue</b> is <b>IntPtr.Zero</b>, 
		/// which is similar to NULL.
		/// </remarks>
		public static readonly IntPtr UndefinedHandleValue = IntPtr.Zero;		// similar to NULL
		
		/// <summary>
		/// Enumeration constants used to denote a process output stream.
		/// </summary>
		/// <remarks>
		/// <b>OutputType</b> constants are used in conjunction with <see cref="OutputLine"/> events.<br/>
		/// Refer to that topic for details on how <b>OutputType</b> is used.
		/// </remarks>
		/// <seealso cref="OutputLine">OutputLine</seealso>
		/// <seealso cref="OutputLineEventArgs">OutputLineEventArgs</seealso>
		/// <seealso cref="OutputLineEventArgs.OutputType">OutputLineEventArgs.OutputType</seealso>
		public enum OutputType
		{
			/// <summary>No output stream. Value is 0.</summary>
			None   = 0,
			/// <summary><b>StandardOutput</b> output stream. Value is 1.</summary>
			StdOut = 1,		
			/// <summary><b>StandardError</b> output stream. Value is 2.</summary>
			StdErr = 2		
		}; // OutputType enum
	
		/// <summary>
		/// String constant for the <b>echo off</b> 'DOS' command.
		/// </summary>
		/// <remarks>
		/// This constant is used in the <see cref="DoStart"/> method to switch off 'DOS'-like text
		/// output, if necessary. Its value is <b>"echo off"</b>. See <b>DoStart</b> for details.
		/// </remarks>
		/// <seealso cref="DoStart">DoStart (protected)</seealso>
		protected const string m_cstrEchoOff = "echo off";
	
		/// <summary>
		/// String constant for the <b>exit</b> 'DOS' command.
		/// </summary>
		/// <remarks>
		/// This constant is used in the <see cref="DoStart"/> method to cause the Windows command line
		/// interpreter to quit immediately after execution of the command line, if necessary. 
		/// Its value is <b>"exit"</b>. See <b>DoStart</b> for details.
		/// </remarks>
		/// <seealso cref="DoStart">DoStart (protected)</seealso>
		protected const string m_cstrExit = "exit";
												
	#endregion // Constants and Enumerations

	#region Properties

		/// <summary>
		/// Gets the arguments portion of the command line.
		/// </summary>
		/// <value>
		/// The arguments portion of the command line (i.e. the command parameters).
		/// <c>Null</c> if there is none.
		/// </value>
		/// <event cref="CmdLineChanged">
		/// The command line has changed.
		/// </event>
		/// <remarks>
		/// The value of <b>Arguments</b> cannot be changed during process execution.
		/// <para/>
		/// Note that changing the <see cref="Command"/> property will automatically set <b>Arguments</b>
		/// to <c>null</c>.
		/// <para/>
		/// If <b>Arguments</b> has changed, a <seealso cref="CmdLineChanged"/> event will be fired.
		/// </remarks>
		/// <seealso cref="m_strArguments">m_strArguments (protected)</seealso>
		/// <seealso cref="CommandLine"/>
		/// <seealso cref="Command"/>
		/// <seealso cref="CmdLineChanged"/>
		public string Arguments
		{
			get
			{
				return m_strArguments;
			}

			set
			{
				if (IsRunning) return;	// don't allow changes while process is running

				if (Command == null) return;	// don't allow setting args without a command being set

				if (value == m_strArguments) return;	// don't set the same value twice

				if ((m_strArguments == null) && (value != null))
					if (value.Trim() == "") return; 	// null and "" are treated the same here

				if ((value != null) && (m_strArguments != null))	// sic
				{
					if (value.Trim().ToUpper() == 								
						m_strArguments.Trim().ToUpper()) return;	// ignore case
				}

				if (value != null)
				{
					if (value.Trim().Length > 0) m_strArguments = value.Trim();	// you never know...
						else  m_strArguments = null;
				}
				else 
				{
					m_strArguments = null;
				}
					
				RaiseCmdLineChangedEvent();	// notify anybody who may listen
			}
		} // Arguments property

		/// <summary>
		/// Gets or sets the command portion of the command line.
		/// </summary>
		/// <value>
		/// The command portion of the command line (i.e. the exe filename without arguments).
		/// <c>Null</c> if there is none.
		/// </value>
		/// <event cref="CmdLineChanged">
		/// The command line has changed.
		/// </event>
		/// <remarks>
		/// The value of <b>Command</b> cannot be changed during process execution.
		/// <para/>
		/// If <b>Command</b> is changed, the <see cref="Arguments"/> property will automatically 
		/// be set to <c>null</c>, the <see cref="UseComSpec"/> property will be set to <c>false</c>,
		/// and a <see cref="CmdLineChanged"/> event will be fired.
		/// </remarks>
		/// <seealso cref="m_strCommand">m_strCommand (protected)</seealso>
		/// <seealso cref="CommandLine"/>
		/// <seealso cref="Arguments"/>
		/// <seealso cref="UseComSpec"/>
		/// <seealso cref="CmdLineChanged"/>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public string Command
		{
			get
			{
				return m_strCommand;
			}

			set
			{
				if (IsRunning) return;	// don't allow changes while process is running

				if (value == m_strCommand) return;	// don't set the same value twice

				if ((m_strCommand == null) && (value != null))
					if (value.Trim() == "") return; 	// null and "" are treated the same here

				if ((value != null) && (m_strCommand != null))		// sic
				{
					if (value.Trim().ToUpper() ==							
						m_strCommand.Trim().ToUpper()) return;		// ignore case
				}

				if (value != null)
				{
					if (value.Trim().Length > 0) m_strCommand = value.Trim();	// you never know...
						else  m_strCommand = null;
				}
				else 
				{
					m_strCommand = null;
				}
					
				m_strArguments = null;		// clear args when setting a new command
				m_bUseComSpec = false;		// assume not to use comspec for a new command

				RaiseCmdLineChangedEvent();	// notify anybody who may listen
			}
		} // Command property

		/// <summary>
		/// Gets the full command line.
		/// </summary>
		/// <value>
		/// The full command line text, which is the concatenation of the <see cref="Command"/>
		/// and <see cref="Arguments"/> property values. <c>Null</c> indicates an empty command line.
		/// </value>
		/// <remarks>
		/// <b>CommandLine</b> returns the command text that as it would be executed by the 
		/// <see cref="Start">Start</see> method when called.
		/// </remarks>
		/// <seealso cref="Command"/>
		/// <seealso cref="Arguments"/>
		/// <seealso cref="Start">Start</seealso>
		public string CommandLine
		{
			get
			{
				if (Command == null) return null;

				string strRet;

				if (Arguments == null) strRet = Command;		// build command line
					else strRet = Command + " " + Arguments;
				
				return strRet;
			}
		} // CommandLine property

		/// <summary>
		/// Returns the default value for the <see cref="OutputCodepage"/> property.
		/// </summary>
		/// <value>
		/// <b>DefaultOutputCodepage</b> returns the value that the <see cref="OutputCodepage"/> property
		/// will have when creating a new instance of <b>CommandLineProcess</b>.
		/// </value>
		/// <remarks>
		/// This implementation of the property returns <b>CultureInfo.CurrentCulture.TextInfo.OEMCodePage</b>.
		/// <para/>
		/// Overwrite this property if you want another value to be set by default.
		/// Use <b>-1</b> if you don't want any re-encoding to be applied to the process output.
		/// </remarks>
		/// <seealso cref = "OutputCodepage"/>
		/// <seealso cref = "CultureInfo.CurrentCulture"/>
		public virtual int DefaultOutputCodepage
		{
			get
			{
				// set OEM code page of current system environment by default
				return CultureInfo.CurrentCulture.TextInfo.OEMCodePage;
			}
		} // DefaultOutputCodepage property

		/// <summary>
		/// Gets the last exit code of the process.
		/// </summary>
		/// <value>
		/// <b>ExitCode</b> is an integer reflecting the last return value of the recently executed
		/// application. <see cref="UndefinedCodeValue"/> denotes an error condition.
		/// </value>
		/// <remarks>
		/// If no application was ever executed or <see cref="Close"/> was called after the last execution,
		/// <b>ExitCode</b> will return <b>UndefinedCodeValue</b>.
		/// <para/>
		/// <note>
		/// Usually an application returns <b>0</b> to signal successful execution. But this is only
		/// convention, not a standard that can blindly be relied on.
		/// </note>
		/// This property mirrors the <see cref="System.Diagnostics.Process.ExitCode">
		/// Process.ExitCode</see> property of the underlying <see cref="System.Diagnostics.Process">
		/// Process</see> component.
		/// </remarks>
		/// <seealso cref="UndefinedCodeValue"/>
		/// <seealso cref="Close"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitCode">System.Diagnostics.Process.ExitCode</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public int ExitCode
		{
			get
			{
				try
				{
					return m_proc.ExitCode;
				}
				catch
				{
					return UndefinedCodeValue; // failure
				}
			}
		} // ExitCode property

		/// <summary>
		/// Gets the last retrieved exit time of the process.
		/// </summary>
		/// <value>
		/// <b>ExitTime</b> retrieves the last exit time that an executed application returned.
		/// <see cref="UndefinedTimeValue"/> stands for 'no time available'.
		/// </value>
		/// <remarks>
		/// If no application was executed or <see cref="Close"/> was called after the last execution,
		/// <b>ExitTime</b> will return <b>UndefinedTimeValue</b>.
		/// <para/>
		/// This property mirrors the <see cref="System.Diagnostics.Process.ExitTime">
		/// Process.ExitTime</see> property of the underlying <see cref="System.Diagnostics.Process">
		/// Process</see> component.
		/// </remarks>
		/// <seealso cref="UndefinedTimeValue"/>
		/// <seealso cref="Close"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitTime">System.Diagnostics.Process.ExitTime</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public DateTime ExitTime 
		{
			get
			{
				try
				{
					return m_proc.ExitTime;
				}
				catch
				{
					return UndefinedTimeValue; // failure
				}
			}
		} // ExitTime property

		/// <summary>
		/// Gets the process' handle.
		/// </summary>
		/// <value>
		/// <b>Handle</b> retrieves the handle value that the operating system assigned to the process
		/// on last execution. <see cref="UndefinedHandleValue"/> denotes an error condition.
		/// </value>
		/// <remarks>
		/// If no application was ever executed or <see cref="Close"/> was called after the last execution,
		/// <b>Handle</b> will return <b>UndefinedHandleValue</b>.
		/// <para/>
		/// <note>
		/// A process handle is private to an application. If you need a system-unique identifier, use
		/// the <see cref="Id"/> property instead.
		/// </note>
		/// <para/>
		/// This property mirrors the  <see cref="System.Diagnostics.Process.Handle">Handle</see>
		/// property of the underlying <see cref="System.Diagnostics.Process">Process</see> component.
		/// </remarks>
		/// <seealso cref="UndefinedHandleValue"/>
		/// <seealso cref="Id"/>
		/// <seealso cref="Close"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.Handle">System.Diagnostics.Process.Handle</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public IntPtr Handle
		{
			get
			{
				try
				{
					return m_proc.Handle;
				}
				catch
				{
					return UndefinedHandleValue; // failure
				}
			}
		} // Handle property

		/// <summary>
		/// Gets a value indicating whether the process has terminated.
		/// </summary>
		/// <value>
		/// <c>True</c> if the process has terminated, otherwise <c>false</c>.
		/// </value>
		/// <remarks>
		/// A value of <c>true</c> for <b>HasExited</b> indicates that the process has terminated, either 
		/// normally or abnormally. You can force the process to exit by calling <see cref="Abort"/>.
		/// <para/>
		/// This property mirrors the  <see cref="System.Diagnostics.Process.HasExited">HasExited</see>
		/// property of the underlying <see cref="System.Diagnostics.Process">Process</see> component.
		/// </remarks>
		/// <seealso cref="Abort"/>
		/// <seealso cref="IsRunning"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.HasExited">System.Diagnostics.Process.HasExited</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public bool HasExited 
		{
			get
			{
				try
				{
					return m_proc.HasExited;
				}
				catch
				{
					return false; // failure
				}
			}
		} // HasExited property

		/// <summary>
		/// Gets the unique identifier of the process.
		/// </summary>
		/// <value>
		/// The system-generated unique identifier of the process. 0 (zero) if the process Id
		/// is not available.
		/// </value>
		/// <remarks>
		/// The process Id is guaranteed to be unique throughout the system. However, it can be reused
		/// by the operating system after the process has exited.
		/// <para/>
		/// This property mirrors the  <see cref="System.Diagnostics.Process.Id">Id</see>
		/// property of the underlying <see cref="System.Diagnostics.Process">Process</see> component.
		/// </remarks>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.Id">System.Diagnostics.Process.Id</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public int Id
		{
			get
			{
				try
				{
					return m_proc.Id;
				}
				catch
				{
					return 0; // failure
				}
			}
		} // Id property

		/// <summary>
		/// Gets a value indicating whether the process is currently executing.
		/// </summary>
		/// <value>
		/// <c>True</c> if the process is currently running, otherwise <c>false</c>.
		/// </value>
		/// <remarks>
		/// You can force the process to exit by calling <see cref="Abort"/>.
		/// <para/>
		/// </remarks>
		/// <seealso cref="Abort"/>
		/// <seealso cref="HasExited"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		public bool IsRunning
		{
			get
			{
				return (!HasExited && (Id != 0));
			}
		} // IsRunning property

		/// <summary>
		/// Gets or sets the code page used to re-encode process output.
		/// </summary>
		/// <value>
		/// The identifier of the code page to use for re-encoding process output. <c>-1</c>
		/// stands for no re-encoding.
		/// </value>
		/// <remarks>
		/// Re-encoding of text takes place after reading the textual output from the running
		/// process component and before raising the <see cref="OutputLine"/> event. The argument
		/// class sent with this event will contain one re-encoded line of text.
		/// <para/>
		/// The value of <b>OutputCodepage</b> cannot be changed during process execution. Also an
		/// invalid code page identifier will be ignored. 
		/// <para/>
		/// To re-encode output text is useful in displaying language-specific special characters 
		/// appropriately (e.g. the german umlaut characters).
		/// <note>
		/// The list of available code pages is stored under the registry key 
		/// <b>HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage</b>.
		/// </note>
		/// </remarks>
		/// <seealso cref="DefaultOutputCodepage"/>
		/// <seealso cref="OutputLine"/>
		/// <seealso cref="ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		/// <seealso cref = "m_encOutput">m_encOutput (protected)</seealso>
		/// <seealso cref="System.Text.Encoding">System.Text.Encoding</seealso>
		/// <seealso cref="System.Text.Encoding.CodePage">System.Text.Encoding.CodePage</seealso>
		public int OutputCodepage
		{
			get
			{
				if (m_encOutput == null) return -1; // -1 means no encoding
				return m_encOutput.CodePage;
			}

			set
			{
				if (IsRunning) return;				// don't allow change while process is running
				if (value < 0)						// no encoding, delete m_encOutput
				{
					if (m_encOutput == null) return;
					m_encOutput = null;
				}
				else
				{
					Encoding e;

					// first check if code page number is supported on this system
					try
					{
						e = Encoding.GetEncoding(value);	// throws an exception if code page doesn't exist
					}
					catch
					{
						return; // code page not supported, leave everything unchanged
					}

					m_encOutput = e;	// set new code page
				}
			}
		} // OutputCodePage property

		/// <summary>
		/// Gets the name of the last executed process.
		/// </summary>
		/// <value>
		/// The name that the system uses to identify the process to the user. <c>Null</c> if no name 
		/// is available.
		/// </value>
		/// <remarks>
		/// The <b>ProcessName</b> property holds an executable file name, (such as e.g. <c>ping</c>
		/// or <c>imp</c>), that usually does not include the <c>.exe</c> extension or the path.
		/// <para/>
		/// If no application was ever executed or <see cref="Close"/> was called after the last execution,
		/// <b>ProcessName</b> will return <c>null</c>.
		/// <para/>
		/// If the <see cref="UseComSpec"/> is set to <c>true</c>, <b>ProcessName</b> consequently will 
		/// always return <c>cmd</c>.
		/// <para/>
		/// This property mirrors the  <see cref="System.Diagnostics.Process.ProcessName">ProcessName</see>
		/// property of the underlying <see cref="System.Diagnostics.Process">Process</see> component.
		/// </remarks>
		/// <seealso cref="Close"/>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		/// <seealso cref="UseComSpec"/>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.ProcessName">System.Diagnostics.Process.ProcessName</seealso>
		public string ProcessName
		{
			get
			{
				try
				{
					return m_proc.ProcessName;
				}
				catch
				{
					return null; // failure
				}
			}
		} // ProcessName property

		/// <summary>
		/// Gets the last start time of the process.
		/// </summary>
		/// <value>
		/// A <see cref="DateTime"/> that indicates when the process started. 
		/// </value>
		/// <remarks>
		/// If no application was ever executed or <see cref="Close"/> was called after the last execution,
		/// <b>StartTime</b> will return <see cref="UndefinedTimeValue"/>.
		/// This property mirrors the <see cref="System.Diagnostics.Process.StartTime">
		/// StartTime</see> property of the underlying <see cref="System.Diagnostics.Process">
		/// Process</see> component.
		/// </remarks>
		/// <seealso cref="Close"/>
		/// <seealso cref="UndefinedTimeValue"/>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		/// <seealso cref="System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso cref="System.Diagnostics.Process.StartTime">System.Diagnostics.Process.StartTime</seealso>
		public DateTime StartTime 
		{
			get
			{
				try
				{
					return m_proc.StartTime;
				}
				catch
				{
					return UndefinedTimeValue; // failure
				}
			}
		} // StartTime property

		/// <summary>
		/// Gets a value indicating whether to use the Windows command line processor.
		/// </summary>
		/// <value>
		/// <b>UseComSpec</b> determines if the current command line - specified by the <see cref="Command"/>
		/// and <see cref="Arguments"/> properties - points to a stand-alone application (*.exe file) or
		/// defines a 'DOS command'.
		/// </value>
		/// <event cref="CmdLineChanged">
		/// The command line has changed.
		/// </event>
		/// <remarks>
		/// If <b>UseComSpec</b> is set to <c>true</c>, <b>CommandLineProcess</b> will pass the current 
		/// command line to the Windows command line processor (called Comspec here) for execution rather
		/// than trying to execute it directly. <br/>
		/// In other words: the current command line will be treated as a 'DOS command'.
		/// For example, <c>dir C:\</c> will be actually executed as <c>cmd dir c:\</c>
		/// when <b>UseComSpec</b> is set to <c>true</c>. Consequently, you will always get
		/// <c>cmd</c> as <see cref="ProcessName"/> when using Comspec.
		/// <para/>
		/// Executing a command via <c>cmd.exe</c> differs in various ways from simply calling another *.exe
		/// file with some additional arguments. Some special actions have to be performed when starting,
		/// aborting and processing output in such a case. See <see cref="DoStart"/>,  <see cref="Abort"/>
		/// and <see cref="ReadProcessOutput"/> for details.
		/// <para/>
		/// The value of <b>UseComSpec</b> cannot be changed during process execution.
		/// <para/>
		/// Note that changing the <see cref="Command"/> property will automatically set <b>UseComSpec</b>
		/// to <c>false</c>.
		/// <para/>
		/// If <b>UseComSpec</b> has changed, a <seealso cref="CmdLineChanged"/> event will be fired.
		/// </remarks>
		/// <seealso cref="m_bUseComSpec">m_bUseComSpec (protected)</seealso>
		/// <seealso cref="Command"/>
		/// <seealso cref="Arguments"/>
		/// <seealso cref="CmdLineChanged"/>
		/// <seealso cref="ProcessName"/>
		/// <seealso cref="DoStart"/>
		/// <seealso cref="Abort"/>
		/// <seealso cref="ReadProcessOutput"/>
		public bool UseComSpec
		{
			get
			{
				return m_bUseComSpec;
			}

			set
			{
				if (IsRunning) return;				// don't allow changes while process is running
				if (Command == null) return;		// don't allow setting value without a command being set
				if (value == m_bUseComSpec) return;	// don't set the same value twice

				m_bUseComSpec = value;

				RaiseCmdLineChangedEvent();	// notify anybody who may listen
			}
		} // UseComSpec property


	#endregion // Properties

	#region Events

		/// <summary>
		/// Signals that the process was started.
		/// </summary>
		/// <remarks>
		/// A <b>Started</b> event is fired when a <b>CommandLineProcess</b> instance successfully called the
		/// <see href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</see> method with the currently set command line. 
		/// The <see cref = "StartedEventArgs.StartTime">StartTime</see> member of the <b>StartedEventArgs</b>
		/// event argument class denotes the time when the starting took place (This value can also be
		/// retrieved via <b>CommandLineProcess' </b> <see cref = "StartTime"/> property.).<br/>
		/// If the starting attempt failed, no event will be raised.<br/>
		/// Refer to the description of the 
		/// <see href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
		/// Start</see> methods for further details.
		/// </remarks>
		/// <seealso href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
		/// Start Overload List</seealso>
		/// <seealso cref = "StartedEventHandler"/>
		/// <seealso cref = "StartedEventArgs"/>
		/// <seealso cref = "StartTime"/>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public event StartedEventHandler Started;
		/// <summary>
		/// Signals that the process has exited.
		/// </summary>
		/// <remarks>
		/// An <b>Exited</b> event is raised when a <b>CommandLineProcess</b> instance receives a
		/// <see cref = "System.Diagnostics.Process.Exited">System.Diagnostics.Process.Exited</see> 
		/// event from the executed command line process. The <see cref = "ExitedEventArgs.ExitCode">
		/// ExitCode</see> and <see cref = "ExitedEventArgs.ExitTime">ExitTime</see> members of the 
		/// <b>ExitedEventArgs</b> argument class denote the exit code and time as retrieved
		/// from the process.<br/> 
		/// <para/>
		/// The reason for process termination can be either normal (i.e. the command line application has
		/// successfully finished) or abnormal (i.e. the command line application was aborted by the system
		/// due to some irregular condition or a call to <see cref = "Abort"/>). In the latter case (<b>Abort</b> 
		/// has been called) an <see cref = "Aborted"/> event will occur prior to the <b>Exited</b> event.
		/// </remarks>
		/// <seealso cref = "ExitedEventHandler"/>
		/// <seealso cref = "ExitedEventArgs"/>
		/// <seealso cref = "OnProcessExited"/>
		/// <seealso cref = "Abort"/>
		/// <seealso cref = "Aborted"/>
		/// <seealso cref = "System.Diagnostics.Process.Exited">System.Diagnostics.Process.Exited</seealso>
		/// <seealso cref = "System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</seealso>
		public event ExitedEventHandler Exited;
		/// <summary>
		/// Signals abortion of the process.
		/// </summary>
		/// <remarks>
		/// An <b>Aborted</b> event is fired when a <b>CommandLineProcess</b> instance successfully called the
		/// <see cref = "System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</see> method for
		/// the running command line process. The <see cref = "AbortedEventArgs.AbortTime">AbortTime</see> 
		/// member of the <b>AbortedEventArgs</b> event argument class denotes the time when the aborting 
		/// attempt happened.<br/>
		/// Successful abortion of the process will result in an additional <see cref = "Exited"/> event.<br/>
		/// No event will be raised in case of failure.<br/>
		/// Refer to the description of the <see cref = "Abort"/> method for further details.
		/// </remarks>
		/// <seealso cref = "Abort"/>
		/// <seealso cref = "AbortedEventHandler"/>
		/// <seealso cref = "AbortedEventArgs"/>
		/// <seealso cref = "Exited"/>
		/// <seealso cref = "System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</seealso>
		/// <seealso cref = "System.Diagnostics.Process.Exited">System.Diagnostics.Process.Exited</seealso>
		public event AbortedEventHandler Aborted;
		/// <summary>
		/// Signals that the command line has changed.
		/// </summary>
		/// <remarks>
		/// <b>CmdLineChanged</b> is fired every time the command line of the <b>CommandLineProcess</b> 
		/// instance has changed by using its <see cref="Command"/>, <see cref="Arguments"/> or
		/// <see cref="UseComSpec"/> property. The <see cref="CmdLineChangedEventArgs.Text">Text</see> 
		/// member of the <b>CmdLineChangedEventArgs</b> event argument class will contain the new command
		/// line text. The <see cref = "CmdLineChangedEventArgs.UseComSpec">UseComSpec</see>
		/// field will indicate whether this command is treated as a stand-alone command line or
		/// as a 'DOS command'.<br/>
		/// No event will be raised if the command line remains unchanged for some reason (e.g.
		/// because the process is currently running).
		/// </remarks>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso cref = "CommandLine"/>
		/// <seealso cref = "CmdLineChangedEventHandler"/>
		/// <seealso cref = "CmdLineChangedEventArgs"/>
		public event CmdLineChangedEventHandler CmdLineChanged;
		/// <summary>
		/// Signals that the system resources for the process were freed.
		/// </summary>
		/// <remarks>
		/// A <b>Closed</b> event is fired when a <b>CommandLineProcess</b> instance successfully called the
		/// <see cref = "System.Diagnostics.Process.Close">System.Diagnostics.Process.Close</see> method
		/// for the system process. This causes the operating system to free all resources that might have
		/// been allocated for this process. Consequently, properties that rely on corresponding properties
		/// of the underlying <see cref="System.Diagnostics.Process">System.Diagnostics.Process</see> object 
		/// will no longer return any valid result after the process has been closed. The 
		/// <see cref = "ClosedEventArgs.CloseTime">CloseTime</see> member of the <b>ClosedEventArgs</b> 
		/// event argument class denotes the time when the closing happened.<br/>
		/// No event will be raised in case of failure.
		/// </remarks>
		/// <seealso cref = "Close"/>
		/// <seealso cref = "ClosedEventHandler"/>
		/// <seealso cref = "ClosedEventArgs"/>
		/// <seealso cref = "System.Diagnostics.Process.Close">System.Diagnostics.Process.Close</seealso>
		public event ClosedEventHandler Closed;
		/// <summary>
		/// Holds one line of textual output from the executing process.
		/// </summary>
		/// <remarks>
		/// An <b>OutputLine</b> event is fired for every line of text produced by the executed command line 
		/// application. The <see cref = "OutputLineEventArgs.Text">Text</see> member of the 
		/// <b>OutputLineEventArgs</b>  event argument class will contain the text of the line (may be <c>null</c>
		/// or <c>""</c>), the value of the <see cref = "OutputLineEventArgs.OutputType">OutputType</see> 
		/// member contains the information whether this text was written to the <b>StandardOutput</b> or 
		/// <b>StandardError</b> stream by the application.
		/// <para/>
		/// The original text might be re-encoded according to the value of the
		/// <see cref="OutputCodepage"/> property.
		/// <note>
		/// Seen from the perspective of an application that is using a <b>CommandLineProcess</b> object to
		/// execute an external command line tool, processing an <b>OutputLine</b> event is a call from another
		/// thread. Thus, the code processing the <b>OutputLine</b> event has to be safe for multi-threading.
		/// </note>
		/// </remarks>
		/// <seealso cref = "OutputLineEventHandler"/>
		/// <seealso cref = "OutputLineEventArgs"/>
		/// <seealso cref = "OutputCodepage"/>
		/// <seealso cref = "ReadProcessOutput"/>
		/// <seealso cref = "System.Diagnostics.Process.StandardOutput">System.Diagnostics.Process.StandardOutput</seealso>
		/// <seealso cref = "System.Diagnostics.Process.StandardError">System.Diagnostics.Process.StandardError</seealso>
		public event OutputLineEventHandler OutputLine;

	#endregion // Events

	#region Operations (public methods)

		/// <overloads>
		/// Starts a command line application, optionally sets the command line related values before.
		/// </overloads>
		/// <summary>
		/// Executes the command line as it is currently set.
		/// </summary>
		/// <event cref = "Started">
		/// Indirect. Indicates that the process has started (actually fired by <see cref = "DoStart"/> 
		/// protected method.)
		/// </event>
		/// <remarks>
		/// Attempts to execute the current command line text as set by using the <see cref = "Command"/>, 
		/// <see cref="Arguments"/> and/or <see cref="UseComSpec"/> properties or an appropriate 
		/// constructor version.<br/>
		/// If the process is already running, the call is ignored and <b>Start</b> returns <c>false</c>.
		/// </remarks>
		/// <returns>
		/// <c>True</c> if the process has started successfully, <c>false</c> otherwise.
		/// </returns>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "DoStart">DoStart (protected)</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public bool Start()
		{
			if (IsRunning) return false;		// Don't try to start twice.
			if (Command == null) return false;	// Don't try to start with empty command line

			return DoStart(); // start process and output recording
		} // Start()

		/// <summary>
		/// Sets the instance's <see cref = "Command"/> property and executes the command.
		/// </summary>
		/// <event cref = "CmdLineChanged">
		/// Indicates that the command line has changed.
		/// </event>
		/// <event cref = "Started">
		/// Indirect. Indicates that the process has started (actually fired by <see cref = "DoStart"/> 
		/// protected method).
		/// </event>
		/// <param name="command">
		/// The command to execute.
		/// </param>
		/// <remarks>
		/// This version of <b>Start</b> first sets the <see cref = "Command"/> property to the value
		/// of the <b>command</b> parameter, then starts a process to execute the command.<br/>
		/// If the function succeeds, first a <see cref = "CmdLineChanged"/> and then a
		/// <see cref = "Started"/> event will be fired.<br/>
		/// If the process is already running, the call is ignored and <b>Start</b> returns <c>false</c>.
		/// <para/>
		/// Note that changing the <b>Command</b> property causes <see cref="Arguments"/> to be set to
		/// <c>null</c> and <see cref="UseComSpec"/> to be set to <c>false</c>.
		/// </remarks>
		/// <returns>
		/// <c>True</c> if the process has started successfully, <c>false</c> otherwise.
		/// </returns>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "CmdLineChanged"/>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "DoStart">DoStart (protected)</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public bool Start(string command)
		{
			if (IsRunning) return false;	// Don't try to start twice.

			// set new command (causes firing of a CmdLineChanged event, sets 
			// Arguments to null and UseComSpec to false)
			this.Command = command;						
			if (this.Command == null) return false;	// Don't try to start with empty command line

			return DoStart(); // start process and output recording
		} // Start(string)

		/// <summary>
		/// Sets the instance's <see cref = "Command"/> and <see cref = "UseComSpec"/>
		/// properties, then executes the new command line.
		/// </summary>
		/// <event cref = "CmdLineChanged">
		/// Indicates that the command line has changed.
		/// </event>
		/// <event cref = "Started">
		/// Indirect. Indicates that the process has started (actually fired by <see cref = "DoStart"/> 
		/// protected method).
		/// </event>
		/// <param name="command">
		/// The command to execute.
		/// </param>
		/// <param name="useComSpec">
		/// Whether or not the Windows command line processor shall be used to execute the command line
		/// ('DOS command').
		/// </param>
		/// <remarks>
		/// This version of <b>Start</b> first sets the <see cref = "Command"/> and
		/// <see cref = "UseComSpec"/> properties to the specified parameter values, then starts a 
		/// process to execute the command line.<br/>
		/// If the function succeeds, first a <see cref = "CmdLineChanged"/> event and then a 
		/// <see cref = "Started"/> event will be fired. If the process is already running,
		/// the call is ignored and <b>Start</b> returns <c>false</c>.
		/// </remarks>
		/// <returns>
		/// <c>True</c> if the process has started successfully, <c>false</c> otherwise.
		/// </returns>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso cref = "CmdLineChanged"/>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "DoStart">DoStart (protected)</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public bool Start(string command, bool useComSpec)
		{
			if (IsRunning) return false;	// Don't try to start twice.

			// empty command and exit ?
			if ((command == null) || (command.Trim() == ""))
			{
				this.Command = null;	// causes firing of a CmdLineChanged event, sets 
				return false;			// Arguments to null and UseComSpec to false
			}

			// set new values and raise a CmdLineChanged event
			m_strCommand = command;		
			m_bUseComSpec = useComSpec;	
			RaiseCmdLineChangedEvent();

			return DoStart(); // start process and output recording
		} // Start(string, bool)

		/// <summary>
		/// Sets the instance's <see cref = "Command"/> and <see cref = "Arguments"/> properties 
		/// and executes the new command line.
		/// </summary>
		/// <event cref = "CmdLineChanged">
		/// Indicates that the command line has changed.
		/// </event>
		/// <event cref = "Started">
		/// Indirect. Indicates that the process has started (actually fired by <see cref = "DoStart"/> 
		/// protected method).
		/// </event>
		/// <param name="command">
		/// The command to execute.
		/// </param>
		/// <param name="arguments">
		/// The command line arguments.
		/// </param>
		/// <remarks>
		/// This version of <b>Start</b> first sets the <see cref = "Command"/> and <see cref = "Arguments"/>
		/// properties to the values of the <b>command</b> and <b>arguments</b> parameters, then starts a 
		/// process to execute the command line.<br/>
		/// If the function succeeds, first a <see cref = "CmdLineChanged"/> event and then a
		///  <see cref = "Started"/> event will be fired.If the process is already running, 
		/// the call is ignored and <b>Start</b> returns <c>false</c>.<br/>
		/// Note that the <see cref="UseComSpec"/> property will automatically be set to <c>false</c>.
		/// </remarks>
		/// <returns>
		/// <c>True</c> if the process has started successfully, <c>false</c> otherwise.
		/// </returns>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "CmdLineChanged"/>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "DoStart">DoStart (protected)</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public bool Start(string command, string arguments)
		{
			if (IsRunning) return false;	// Don't try to start twice.

			// empty command and exit ?
			if ((command == null) || (command.Trim() == ""))
			{
				this.Command = null;	// causes firing of a CmdLineChanged event, sets 
				return false;			// Arguments to null and UseComSpec to false
			}

			// set new values and raise a CmdLineChanged event
			m_strCommand = command;		
			m_strArguments = arguments;	
			m_bUseComSpec = false;		
			RaiseCmdLineChangedEvent();

			return DoStart(); // start process and output recording

		} // Start(string, string)

		/// <summary>
		/// Sets the instance's <see cref = "Command"/>, <see cref = "Arguments"/>, and
		/// <see cref = "UseComSpec"/> properties, then executes the new command line.
		/// </summary>
		/// <event cref = "CmdLineChanged">
		/// Indicates that the command line has changed.
		/// </event>
		/// <event cref = "Started">
		/// Indirect. Indicates that the process has started (actually fired by <see cref = "DoStart"/> 
		/// protected method).
		/// </event>
		/// <param name="command">
		/// The command to execute.
		/// </param>
		/// <param name="arguments">
		/// The command line arguments.
		/// </param>
		/// <param name="useComSpec">
		/// Whether or not the Windows command line processor shall be used to execute the command line
		/// ('DOS command').
		/// </param>
		/// <remarks>
		/// This version of <b>Start</b> first sets the <see cref = "Command"/>, <see cref = "Arguments"/>,
		/// and <see cref = "UseComSpec"/> properties to the specified parameter values, then starts a 
		/// process to execute the command line.<br/>
		/// If the function succeeds, first a <see cref = "CmdLineChanged"/> event and then a 
		/// <see cref = "Started"/> event will be fired. If the process is already running,
		/// the call is ignored and <b>Start</b> returns <c>false</c>.
		/// </remarks>
		/// <returns>
		/// <c>True</c> if the process has started successfully, <c>false</c> otherwise.
		/// </returns>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso cref = "CmdLineChanged"/>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "DoStart">DoStart (protected)</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		public bool Start(string command, string arguments, bool useComSpec)
		{
			if (IsRunning) return false;	// Don't try to start twice.

			// empty command and exit ?
			if ((command == null) || (command.Trim() == ""))
			{
				this.Command = null;	// causes firing of a CmdLineChanged event, sets 
				return false;			// Arguments to null and UseComSpec to false
			}

			// set new values and raise a CmdLineChanged event
			m_strCommand = command;		
			m_strArguments = arguments;	
			m_bUseComSpec = useComSpec;		
			RaiseCmdLineChangedEvent();

			return DoStart(); // start process and output recording

		} // Start(string, string, bool)

		/// <summary>
		/// Immediately aborts the running process.
		/// </summary>
		/// <returns>
		/// <c>True</c> if abortion of the running process was successful, <c>false</c> otherwise.
		/// </returns>
		/// <remarks>
		/// <b>Abort</b> attempts to cancel the execution of the currently running process by calling the
		/// <see cref = "System.Diagnostics.Process.Kill"> Kill</see> method of the <see cref = "m_proc">
		/// m_proc</see> member. A successful call will result in an <see cref = "Aborted"/> event being fired.
		/// <para/>
		/// Terminating the process will result in receiving a <see cref = "System.Diagnostics.Process.Exited">
		/// Process.Exited</see> event from the process via the <see cref = "OnProcessExited"/> event handler. 
		/// This then results in a <see cref = "Exited">CommandLineProcess.Exited</see> event being fired
		/// subsequent to the <b>Aborted</b> event. If there is some pending text output in the process' pipes
		/// at the time of abortion, this output is broadcasted by one or more <see cref = "OutputLine"/>
		/// events before firing the conclusive <b>Exited</b> event.
		/// <para/>
		/// The way <b>Abort</b> executes depends on the value of the <see cref="UseComSpec"/> property,
		/// which determines whether or not the Windows command line processor is invoked to execute the
		/// command line. The internal behaviour of the method is totally different for the two
		/// possibilities. However, to the caller, this difference is fully transparent.<br/>
		/// For <b>UseComSpec</b> set to <c>false</c>, the system process represented by the <b>m_proc</b>
		/// member is killed, and that's it.<br/>
		/// For <b>UseComSpec</b> set to <c>true</c>, first a check is made to see whether the command line
		/// processor has started another process to execute the desired command line. This is the case
		/// for some commands (e.g. 'tree'), for others it is not, since they are handled internally (e.g.
		/// 'dir'). If a related additional process can be found, it is killed, causing the command line
		/// processor itself to terminate as well (because there is a pending 'exit' command, see 
		/// <see cref="DoStart"/>). If no such additional process could be found, the command line
		/// processor has to be killed directly, leading to the same result.
		/// </remarks>
		/// <event cref = "Aborted">
		/// Signals abortion of the process.
		/// </event>
		/// <event cref = "OutputLine">
		/// Indirect. Eventually. Fired to broadcast pending text output, if necessary (then actually 
		/// fired by <see cref = "ReadProcessOutput"/> protected method).
		/// </event>
		/// <event cref = "Exited">
		/// Indirect. Signals that the process actually has exited (actually fired by 
		/// <see cref = "OnProcessExited"/> protected method).
		/// </event>
		/// <seealso cref = "Aborted"/>
		/// <seealso cref = "Exited">CommandLineProcess.Exited</seealso>
		/// <seealso cref = "ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		/// <seealso cref = "OnProcessExited">OnProcessExited (protected)</seealso>
		/// <seealso cref = "OutputLine"/>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "System.Diagnostics.Process.Kill">System.Diagnostics.Process.Kill</seealso>
		/// <seealso cref = "Exited">System.Diagnostics.Process.Exited</seealso>
		public bool Abort()
		{
			try
			{
				if (m_bUseComSpec)	// We are running a 'DOS command' ...
				{
					// First check if there is a process running that has the name of our 'DOS command'.
					// This may or may not be true, because some of the 'DOS commands' are actually executed
					// by external processes whereas others are not. E.g. the 'tree' command is actually 
					// executed by calling the 'tree.com' executable, whereas the 'dir' command is handled
					// internally by the command line interpreter.
					Process[] procs = Process.GetProcessesByName(m_strCommand);

					int nCount = procs.GetLength(0);
					Process procToKill = null;
					int nCmdLen = m_strCommand.Length;

					if (nCount == 1) // We have found one (and only one) process that has the name we look for.
					{
						procToKill = procs[0];
					}
					else if (nCount == 0)	// No such process, but there may be one that has a file extension
					{						// appended to its name. E.g. executing the 'tree' command raises
											// a process identified to the system as 'tree.com'.
						foreach (Process p in Process.GetProcesses())
						{
							if ((p.StartTime > m_proc.StartTime) && (p.ProcessName.Length > nCmdLen))
							{
								if (p.ProcessName.Substring(0, nCmdLen) == m_strCommand)
								{
									if (procToKill == null) procToKill = p;
									if (p.StartTime < procToKill.StartTime) procToKill = p;
								}
							}
						}
					}
					else	// We have found more than one process of the desired name. In this case,
					{		// we kill the one which was started immediately after the command line
							// interpreter itself (which is represented by the m_proc variable).
						foreach (Process p in procs)
						{
							if (p.StartTime > m_proc.StartTime)
							{
								if (procToKill == null) procToKill = p;
								if (p.StartTime < procToKill.StartTime) procToKill = p;
							}
						}
					}

					// If we have found an additional system process that was started by the process
					// represented by the m_proc variable (which is actually an instance of the command 
					// line processor), we kill it here and this will cause m_proc to terminate as well.
					// If no additional process was found, we have to kill m_proc directly.
					if (procToKill != null) procToKill.Kill();
						else m_proc.Kill();
				}
				else 
				{
					m_proc.Kill();
				}
			}
			catch
			{
				return false;	// failure
			}

			RaiseAbortedEvent(DateTime.Now);	// notify anybody who may listen
			return true;
		} // Abort()

		/// <summary>
		/// Frees all the resources that are associated with the process.
		/// </summary>
		/// <returns>
		/// <c>True</c> if successful, <c>false</c> otherwise (e.g. if the process is currently executing).
		/// </returns>
		/// <remarks>
		/// <b>Close</b> calls the <see cref = "System.Diagnostics.Process.Close">Close</see> 
		/// method of the <see cref = "m_proc"/> member. A successful call will result in a 
		/// <see cref = "Closed"/> event being fired.
		/// <note>
		/// After a call to <b>Close</b>, the various <b>CommandLineProcess</b> properties, that rely 
		/// on corresponding properties of the <b>m_proc</b> process member, will no longer return
		/// meaningful values.
		/// </note>
		/// <para/>
		/// </remarks>
		/// <event cref = "Closed">
		/// Signals that the process has been closed.
		/// </event>
		/// <seealso cref = "Closed"/>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "System.Diagnostics.Process.Close">System.Diagnostics.Process.Close</seealso>
		public bool Close()
		{
			if (IsRunning) return false;	// Don't call Close() for a running process.
			if (Handle == UndefinedHandleValue)  return false;	// Process should have run at least once.
			try
			{
				m_proc.Close();
			}
			catch
			{
				return false;	// failure
			}

			RaiseClosedEvent(DateTime.Now);	// notify anybody who may listen
			return true;
		} // Close()

	#endregion // Operations (public methods)

	#region Overrides

		/// <summary>
		/// Overridden. Returns a type name for the class.
		/// </summary>
		/// <returns>
		/// A string containing the class name for the object.
		/// </returns>
		/// <remarks>
		/// Returns the string <b>"twComps.CommandLineProcess"</b>.
		/// </remarks>
		public override string ToString()
		{
			return "twComps.CommandLineProcess";
		}


	#endregion // Overrides

	#region Implementation (private, protected, and internal methods)

		/// <summary>
		/// Performs basic initialization.
		/// </summary>
		/// <remarks>
		/// <b>Init</b> is the central method for object initialization. It is called by the various constructor
		/// versions.<br/>
		/// The <see cref = "m_proc"/> and the <see cref = "m_encOutput"/> members are prepared for usage.
		/// <para/>
		/// <notes>
		/// Be sure to call the base class implementation of <b>Init</b> in derived classes.
		/// </notes>
		/// </remarks>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "m_encOutput">m_encOutput (protected)</seealso>
		protected virtual void Init()
		{
			// set Process invariant part of startup info
			m_proc.StartInfo.CreateNoWindow			=			// no (console) window 
			m_proc.StartInfo.RedirectStandardError	=			// Redirect all text
			m_proc.StartInfo.RedirectStandardInput	=			//  streams from and to
			m_proc.StartInfo.RedirectStandardOutput	= true;		//  the process.
			m_proc.StartInfo.UseShellExecute		=			// Don't use the Windows shell.
			m_proc.StartInfo.ErrorDialog			= false;	// Don't display any dialogs.

			// Cause 'Exited' event to be raised on process termination.
			m_proc.EnableRaisingEvents = true;	

			// Wire our Event Handler to process' 'Exited' event,
			m_proc.Exited += new EventHandler(OnProcessExited);

			// set default output encoding
			if (this.DefaultOutputCodepage < 0)
			{
				m_encOutput = null;
			}
			else
			{
				try
				{
					m_encOutput = Encoding.GetEncoding(DefaultOutputCodepage);
				}
				catch
				{
					m_encOutput = null;
				}
			}
		} // Init()

		/// <summary>
		/// Starts the execution of the command line together with a thread intercepting 
		/// the resulting text output.
		/// </summary>
		/// <returns>
		/// <b>True</b> if the process was successfully started, <b>false</b> otherwise.
		/// </returns>
		/// <event cref = "Started">
		/// Indicates that the process has started.
		/// </event>
		/// <remarks>
		/// Calls the <b>Start</b> method of the <see cref="System.Diagnostics.Process">
		/// System.Diagnostics.Process</see> member, on success raises a <b>Started</b> event 
		/// and starts a thread for intercepting process output (by calling <see cref="StartOutputRecording"/>).
		/// <para/>
		/// The way <b>DoStart</b> executes depends on the value of the <see cref="UseComSpec"/> property,
		/// which determines whether or not the Windows command line processor shall be invoked to execute the
		/// command line. The internal behaviour of the method is totally different for the two
		/// possibilities. However, to the caller, this difference is fully transparent.<br/>
		/// For <b>UseComSpec</b> set to <c>false</c>, <b>Command</b> is simply treated as an executable
		/// file to be called with the arguments given in the <b>Arguments</b> property.<br/>
		/// For <b>UseComSpec</b> set to <c>true</c>, first a new instance of the Windows command line 
		/// processor ('an invisible DOS prompt') is started and set to silent mode using the "echo off"
		/// command. After that, a command line composed of the <b>Command</b> and <b>Arguments</b> property
		/// is sent to cmd.exe for execution, immediately followed by an "exit" command, which forces
		/// the cmd.exe process to quit immediately after finishing execution of the previous command.
		/// <para/>
		/// This method is called by the various
		/// <see href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">
		/// Start</see> methods.
		/// </remarks>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "StartOutputRecording">StartOutputRecording (protected)</seealso>
		/// <seealso cref = "System.Diagnostics.Process">System.Diagnostics.Process</seealso>
		/// <seealso href = "ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpref/html/frlrfsystemdiagnosticsprocessclassstarttopic1.htm">
		/// System.Diagnostics.Process.Start</seealso>
		protected virtual bool DoStart()
		{
			if (CommandLine == null) return false;	// nothing to do ...

			try
			{
				if (m_bUseComSpec)
				{
					// set Process start info to call the 'DOS interpreter'
					// (the 'COMSPEC' environment variable points to it)
					m_proc.StartInfo.FileName  = Environment.GetEnvironmentVariable("COMSPEC");
				}
				else
				{
					// set Process start info to execute command line program directly
					m_proc.StartInfo.FileName  = m_strCommand;
					m_proc.StartInfo.Arguments = m_strArguments;
				}

				if (!m_proc.Start()) return false;	// start process
				RaiseStartedEvent(StartTime);		// notify anybody who may listen
				StartOutputRecording();				// start listening thread

				// Some extra actions have to be taken for commands executed by the 'DOS interpreter'
				if (m_bUseComSpec) 
				{
					// first use "echo off" to suppress path prompts ...
					m_proc.StandardInput.WriteLine(m_cstrEchoOff);
					// ... then execute actual command ...
					m_proc.StandardInput.WriteLine(CommandLine);
					// ... and quit the command line processor (cmd.exe) immediately after
					// finishing.
					m_proc.StandardInput.WriteLine(m_cstrExit);
				}
			}
			catch
			{
				return false; // failure
			}
			return true; // success

		} // DoStart()

		/// <summary>
		/// Creates and starts a thread that receives the text output of the process.
		/// </summary>
		/// <returns>
		/// <c>True</c> if the thread was successfully started; <c>false</c> otherwise.
		/// </returns>
		/// <remarks>
		/// <b>StartOutputRecording</b> creates and starts a new <see cref = "System.Threading.Thread">
		/// System.Threading.Thread</see> object (represented by the <b>m_threadRec</b> member).
		/// The thread is wired to the <see cref = "ReadProcessOutput"/> callback function, which
		/// will read the two output streams of the executed process process member.
		/// <para/>
		/// If the function fails, the <see cref = "m_threadRec"/> member will be <c>null</c>.
		/// </remarks>
		/// <seealso cref = "ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		/// <seealso cref = "m_threadRec">m_threadRec (protected)</seealso>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "System.Threading.Thread">System.Threading.Thread</seealso>
		/// <seealso cref = "System.Threading.ThreadStart">ThreadStart Delegate</seealso>
		protected virtual bool StartOutputRecording()
		{
			try
			{
				m_threadRec = new Thread(new ThreadStart(ReadProcessOutput));	// Create listening thread that
				m_threadRec.Name = "CommandLineProcessListener";				//    will execute our callback fn.

				if (m_bUseComSpec) m_bBlockDosOutput = true;	// Causes the first few lines (the 'DOS stuff') of
																//     output being ignored when running comspec.
				m_threadRec.Start();
				return true;
			}
			catch
			{
				m_threadRec = null;
				return false;		// failure
			}
		} // StartOutputRecording()

		/// <summary>
		/// Stops and frees the thread that receives the text output of the process.
		/// </summary>
		/// <remarks>
		/// <b>StopOutputRecording</b> aborts the execution of the thread object represented by
		/// <see cref = "m_threadRec"/> and sets <b>m_threadRec</b> to <c>null</c>.
		/// </remarks>
		/// <seealso cref = "m_threadRec">m_threadRec (protected)</seealso>
		/// <seealso cref = "System.Threading.Thread">System.Threading.Thread</seealso>
		/// <seealso cref = "System.Threading.Thread.Abort">System.Threading.Thread.Abort</seealso>
		protected virtual void StopOutputRecording()
		{
			if (m_threadRec == null) return; // nothing to be done

			try
			{
				m_threadRec.Abort();	// terminate listening thread
			}
			catch {}
			finally 
			{
				m_threadRec = null;	// delete the Thread object in any case
			}
		} // StopOutputRecording()

		/// <summary>
		/// <see cref = "System.Threading.ThreadStart">ThreadStart</see> delegate used
		///  for querying process text output.
		/// </summary>
		/// <event cref = "OutputLine">
		/// Sends one line of process output text.
		/// </event>
		/// <remarks>
		/// <b>ReadProcessOutput</b> is the <see cref = "System.Threading.ThreadStart">ThreadStart</see>
		/// delegate function that is associated with the <see cref = "m_threadRec"/> thread member.
		/// I.e. every time the thread represented by <b>m_threadRec</b> is scheduled for execution
		/// by the operating system, this will result in a call to <b>ReadProcessOutput</b>.
		/// <para/>
		/// The method examines both the  <see cref="System.Diagnostics.Process.StandardOutput">
		/// StandardOutput</see> and the  <see cref="System.Diagnostics.Process.StandardError">
		/// StandardError</see> streams of the process object <see cref = "m_proc"/> to read the
		/// application's output. If some text was read, it is broadcasted via the 
		/// <see cref = "OutputLine"/> event.<br/>
		/// 'DOS'-related text output, including the command line itself, will be internally 
		/// intercepted by this method, if necessary.
		/// <para/>
		/// Before being sent, the text may undergo a re-encoding according to the value of the
		/// <see cref = "OutputCodepage"/> property (<see cref = "m_encOutput"/> and 
		/// <see cref = "m_buffer"/> are used for that purpose).
		/// </remarks>
		/// <seealso cref = "m_threadRec">m_threadRec (protected)</seealso>
		/// <seealso cref = "m_proc">m_proc (protected)</seealso>
		/// <seealso cref = "m_encOutput">m_encOutput (protected)</seealso>
		/// <seealso cref = "m_buffer">m_buffer (protected)</seealso>
		/// <seealso cref = "OutputLine"/>
		/// <seealso cref = "OutputCodepage"/>
		/// <seealso cref = "System.Text.Encoding">System.Text.Encoding</seealso>
		/// <seealso cref = "System.Threading.Thread">System.Threading.Thread</seealso>
		/// <seealso cref = "System.Threading.ThreadStart">ThreadStart Delegate</seealso>
		protected virtual void ReadProcessOutput()
		{
			if (m_threadRec == null) return;	// recording was either stopped or not yet successfully started

			if (OutputLine == null)				// No event handler is wired to our OutputLine event.
			{
				m_proc.StandardOutput.DiscardBufferedData();	// Empty process output streams (process may 
				m_proc.StandardError.DiscardBufferedData();		//  be blocked if stream buffer runs full).
				return;		
			}

			string strLine;
            
			// check process' StandardOutput stream
			try
			{
				while ((strLine = m_proc.StandardOutput.ReadLine()) != null)	
				{
					// prevent the first few lines and the last line from being sent
					// when running a DOS command
					if (m_bUseComSpec)
					{
						if (m_bBlockDosOutput)
						{
							// skip all lines prior to and including the command line (the 'DOS stuff')
							if (strLine == CommandLine) m_bBlockDosOutput = false; 
							continue;
						}
						else
						{
							if (strLine == m_cstrExit) continue; // skip "exit" command
						}
					}

					// convert output text encoding, if necessary
					if (m_encOutput != null)
					{
						m_buffer = m_proc.StandardOutput.CurrentEncoding.GetBytes(strLine);
						RaiseOutputLineEvent(OutputType.StdOut, m_encOutput.GetString(m_buffer));	// send line
					}
					else
					{
						RaiseOutputLineEvent(OutputType.StdOut, strLine);	// send line
					}
				}
			}
			catch {}

			// check process' StandardError stream
			try
			{
				while ((strLine = m_proc.StandardError.ReadLine()) != null)
				{
					// convert output text encoding, if necessary
					if (m_encOutput != null)
					{
						m_buffer = m_proc.StandardError.CurrentEncoding.GetBytes(strLine);
						RaiseOutputLineEvent(OutputType.StdErr, m_encOutput.GetString(m_buffer));	// send line
					}
					else
					{
						RaiseOutputLineEvent(OutputType.StdErr, strLine);	// send line
					}
				}
			}
			catch {}
		} // ReadProcessOutput()

		/// <overloads>
		/// Raises an <see cref="Aborted"/> event with various arguments.
		/// </overloads>
		/// <event cref = "Aborted">
		/// Signals abortion of the process.
		/// </event>
		/// <summary>
		/// Raises an <b>Aborted</b> event with no specified time.
		/// </summary>
		/// <remarks>
		/// The <b>AbortTime</b> member of the <b>AbortedEventArgs</b> event argument class used in
		/// conjunction with this event is set to <b>UndefinedTimeValue</b>.
		/// <para/>
		/// Use a delegate of type <see cref="AbortedEventHandler"/> to register with the <b>Aborted</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Aborted"/>
		/// <seealso cref = "AbortedEventArgs"/>
		/// <seealso cref = "AbortedEventHandler"/>
		/// <seealso cref = "UndefinedTimeValue"/>
		protected void RaiseAbortedEvent()
		{
			if (Aborted != null)
				Aborted(this, new AbortedEventArgs());
		} // RaiseAbortedEvent()

		/// <summary>
		/// Raises an <b>Aborted</b> event and sets the <b>AbortTime</b> argument value.
		/// </summary>
		/// <param name="time">
		/// The time to set for the <b>AbortedEventArgs.AbortTime</b> member.
		/// </param>
		/// <event cref = "Aborted">
		/// Signals abortion of the process.
		/// </event>
		/// <remarks>
		/// The <b>AbortTime</b> member of the <b>AbortedEventArgs</b> event argument class used in
		/// conjunction with this event is set to the value of the <b>time</b> parameter.
		/// <para/>
		/// Use a delegate of type <see cref="AbortedEventHandler"/> to register with the <b>Aborted</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Aborted"/>
		/// <seealso cref = "AbortedEventArgs"/>
		/// <seealso cref = "AbortedEventHandler"/>
		protected void RaiseAbortedEvent(DateTime time)
		{
			if (Aborted != null)
				Aborted(this, new AbortedEventArgs(time));
		} // RaiseAbortedEvent(DateTime)

		/// <overloads>
		/// Raises an <see cref="OutputLine"/> event with or without text.
		/// </overloads>
		/// <summary>
		/// Raises an <b>OutputLine</b> event with the <b>Text</b> member set to <c>null</c>.
		/// </summary>
		/// <param name="type">
		/// Value to set for the <b>OutputLineEventArgs.OutputType</b> member.
		/// </param>
		/// <event cref = "OutputLine">
		/// Holds one line of textual output from the executing process.
		/// </event>
		/// <remarks>
		/// The <b>Text</b> member of the <b>OutputLineEventArgs</b> event argument class used in
		/// conjunction with this event is set to <c>null</c>. <b>OutputType</b> is set to <b>type</b>.
		/// <para/>
		/// Use a delegate of type <see cref="OutputLineEventHandler"/> to register with the <b>OutputLine</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "OutputLine"/>
		/// <seealso cref = "OutputLineEventArgs"/>
		/// <seealso cref = "OutputLineEventHandler"/>
		/// <seealso cref = "OutputType"/>
		protected void RaiseOutputLineEvent(OutputType type)
		{
			if (OutputLine != null)
				OutputLine(this, new OutputLineEventArgs(type));
		} // RaiseOutputLineEvent(CommandLineProcess.OutputType)

		/// <summary>
		/// Raises an <b>OutputLine</b> event and sets the given <b>Text</b> argument.
		/// </summary>
		/// <param name="type">
		/// Value to set for the <b>OutputLineEventArgs.OutputType</b> member.
		/// </param>
		/// <param name="text">
		/// Value to set for the <b>OutputLineEventArgs.Text</b> member.
		/// </param>
		/// <event cref = "OutputLine">
		/// Holds one line of textual output from the executing process.
		/// </event>
		/// <remarks>
		/// The <b>Text</b> member of the <b>OutputLineEventArgs</b> event argument class used in
		/// conjunction with this event is set to <b>text</b>. <b>OutputType</b> is set to <b>type</b>.
		/// <para/>
		/// Use a delegate of type <see cref="OutputLineEventHandler"/> to register with the <b>OutputLine</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "OutputLine"/>
		/// <seealso cref = "OutputLineEventArgs"/>
		/// <seealso cref = "OutputLineEventHandler"/>
		/// <seealso cref = "OutputType"/>
		protected void RaiseOutputLineEvent(OutputType type, string text)
		{
			if (OutputLine != null)
				OutputLine(this, new OutputLineEventArgs(type, text));
		} // RaiseOutputLineEvent(CommandLineProcess.OutputType, text)

		/// <overloads>
		/// Raises a <see cref="Closed"/> event with various arguments.
		/// </overloads>
		/// <event cref = "Closed">
		/// Signalled when the process is closed.
		/// </event>
		/// <summary>
		/// Raises a <b>Closed</b> event with no specified time.
		/// </summary>
		/// <remarks>
		/// The <b>CloseTime</b> member of the <b>ClosedEventArgs</b> event argument class used in
		/// conjunction with this event is set to <b>UndefinedTimeValue</b>.
		/// <para/>
		/// Use a delegate of type <see cref="ClosedEventHandler"/> to register with the <b>Closed</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Closed"/>
		/// <seealso cref = "ClosedEventArgs"/>
		/// <seealso cref = "ClosedEventHandler"/>
		/// <seealso cref = "UndefinedTimeValue"/>
		protected void RaiseClosedEvent()
		{
			if (Closed != null)
				Closed(this, new ClosedEventArgs());
		} // RaiseClosedEvent()

		/// <summary>
		/// Raises a <b>Closed</b> event and sets the <b>CloseTime</b> argument value.
		/// </summary>
		/// <param name="time">
		/// The time to set for the <b>ClosedEventArgs.CloseTime</b> member.
		/// </param>
		/// <event cref = "Closed">
		/// Signalled when the process is closed.
		/// </event>
		/// <remarks>
		/// The <b>CloseTime</b> member of the <b>ClosedEventArgs</b> event argument class used in
		/// conjunction with this event is set to the value of the <b>time</b> parameter.
		/// <para/>
		/// Use a delegate of type <see cref="ClosedEventHandler"/> to register with the <b>Closed</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Closed"/>
		/// <seealso cref = "ClosedEventArgs"/>
		/// <seealso cref = "ClosedEventHandler"/>
		protected void RaiseClosedEvent(DateTime time)
		{
			if (Closed != null)
				Closed(this, new ClosedEventArgs(time));
		} // RaiseClosedEvent(DateTime)

		/// <overloads>
		/// Raises a <see cref="Started"/> event with various arguments.
		/// </overloads>
		/// <summary>
		/// Raises a <b>Started</b> event with no specified time.
		/// </summary>
		/// <event cref = "Started">
		/// Signalled when the process was started.
		/// </event>
		/// <remarks>
		/// The <b>StartTime</b> member of the <b>StartedEventArgs</b> event argument class used in
		/// conjunction with this event is set to <b>UndefinedTimeValue</b>.
		/// <para/>
		/// Use a delegate of type <see cref="StartedEventHandler"/> to register with the <b>Started</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "StartedEventArgs"/>
		/// <seealso cref = "StartedEventHandler"/>
		/// <seealso cref = "UndefinedTimeValue"/>
		protected void RaiseStartedEvent()
		{
			if (Started != null)
				Started(this, new StartedEventArgs());
		} // RaiseStartedEvent()

		/// <summary>
		/// Raises an <b>Started</b> event and sets the <b>StartTime</b> argument value.
		/// </summary>
		/// <param name="time">
		/// The time to set for the <b>StartedEventArgs.StartTime</b> member.
		/// </param>
		/// <event cref = "Started">
		/// Signalled when the process was started.
		/// </event>
		/// <remarks>
		/// The <b>StartTime</b> member of the <b>StartedEventArgs</b> event argument class used in
		/// conjunction with this event is set to the value of the <b>time</b> parameter.
		/// <para/>
		/// Use a delegate of type <see cref="StartedEventHandler"/> to register with the <b>Started</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Started"/>
		/// <seealso cref = "StartedEventArgs"/>
		/// <seealso cref = "StartedEventHandler"/>
		protected void RaiseStartedEvent(DateTime time)
		{
			if (Started != null)
				Started(this, new StartedEventArgs(time));
		} // RaiseStartedEvent(DateTime)

		/// <overloads>
		/// Raises an <see cref="Exited"/> event with various arguments.
		/// </overloads>
		/// <summary>
		/// Raises an <b>Exited</b> event without specified exit time and code.
		/// </summary>
		/// <event cref = "Exited">
		/// Signalled when the process has exited.
		/// </event>
		/// <remarks>
		/// The <b>ExitTime</b> member of the <b>ExitedEventArgs</b> event argument class used in
		/// conjunction with this event is set to <b>UndefinedTimeValue</b>, the <b>ExitCode</b>
		/// member is set to <b>UndefinedCodeValue</b>.
		/// <para/>
		/// Use a delegate of type <see cref="ExitedEventHandler"/> to register with the <b>Exited</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Exited"/>
		/// <seealso cref = "ExitedEventArgs"/>
		/// <seealso cref = "ExitedEventHandler"/>
		/// <seealso cref = "OnProcessExited">OnProcessExited (protected)</seealso>
		/// <seealso cref = "UndefinedTimeValue"/>
		/// <seealso cref = "UndefinedCodeValue"/>
		protected void RaiseExitedEvent()
		{
			if (Exited != null)
				Exited(this, new ExitedEventArgs());
		} // RaiseExitedEvent()

		/// <summary>
		/// Raises an <b>Exited</b> event and sets the <b>ExitTime</b> and <b>ExitCode</b> argument values.
		/// </summary>
		/// <param name="code">
		/// The process' exit code.
		/// </param>
		/// <param name="time">
		/// The time to set for the <b>StartedEventArgs.StartTime</b> member.
		/// </param>
		/// <event cref = "Exited">
		/// Signalled when the process has exited.
		/// </event>
		/// <remarks>
		/// The <b>ExitTime</b> member of the <b>ExitedEventArgs</b> event argument class used in
		/// conjunction with this event is set to the value of the <b>time</b> parameter, the
		/// <b>ExitCode</b> member is set to <b>code</b>.
		/// <para/>
		/// Use a delegate of type <see cref="ExitedEventHandler"/> to register with the <b>Exited</b>
		/// event. The event handler is called whenever the event occurs, unless you remove the delegate. 
		/// For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "Exited"/>
		/// <seealso cref = "ExitedEventArgs"/>
		/// <seealso cref = "ExitedEventHandler"/>
		/// <seealso cref = "OnProcessExited">OnProcessExited (protected)</seealso>
		protected void RaiseExitedEvent(int code, DateTime time)
		{
			if (Exited != null)
				Exited(this, new ExitedEventArgs(code, time));
		} // RaiseExitedEvent(int, DateTime)

		/// <summary>
		/// Raises a <see cref="CmdLineChanged"/> event with the new command line.
		/// </summary>
		/// <event cref = "CmdLineChanged">
		/// Signalled when the objects command line (or its destinated usage) has changed.
		/// </event>
		/// <remarks>
		/// A <b>CmdLineChanged</b> event is raised as a result of using the objects <b>Command</b>,
		/// <b>Arguments</b>, or <b>UseComSpec</b> property. It can also occur as a result of using 
		/// one of the <b>Start</b> methods that expect a parameter.
		/// <para/>
		/// The <b>Text</b> member of the <b>CmdLineChangedEventArgs</b> event argument class used in
		/// conjunction with this event is set to the value that is retrieved by the <b>CommandLine</b> 
		/// property. The <b>UseComSpec</b> field member of the event argument class mirrors the value
		/// of the instance's own <b>UseComSpec</b> property.
		/// <para/>
		/// Use a delegate of type <see cref="CmdLineChangedEventHandler"/> to register with the 
		/// <b>CmdLineChanged</b> event. The event handler is called whenever the event occurs, unless you
		/// remove the delegate. For more information about handling events with delegates, see 
		/// <see href="ms-help://MS.VSCC.2003/MS.MSDNQTR.2003APR.1033/cpguide/html/cpconeventsdelegates.htm">
		/// Events and Delegates</see>.	
		/// </remarks>
		/// <seealso cref = "CmdLineChanged"/>
		/// <seealso cref = "CmdLineChangedEventArgs"/>
		/// <seealso cref = "CmdLineChangedEventHandler"/>
		/// <seealso cref = "CommandLine"/>
		/// <seealso cref = "Command"/>
		/// <seealso cref = "Arguments"/>
		/// <seealso cref = "UseComSpec"/>
		/// <seealso href="mk:@MSITStore:CommandLineProcess.chm::/twComps.CommandLineProcess.Start_overloads.html">Start()</seealso>
		protected void RaiseCmdLineChangedEvent()
		{
			if (CmdLineChanged != null)
				CmdLineChanged(this, new CmdLineChangedEventArgs(CommandLine, UseComSpec));
		} // RaiseCmdLineChangedEvent()

		/// <summary>
		/// Delegate to handle the <see cref="System.Diagnostics.Process.Exited">Exited</see> 
		/// event of the process member.
		/// </summary>
		/// <param name="sender">
		/// The source of the event.
		/// </param>
		/// <param name="e">
		/// An argument of type <see cref="System.EventArgs"/>.
		/// </param>
		/// <event cref = "OutputLine">
		/// Indirect. Fired to broadcast pending text output, if necessary (then actually fired by 
		/// <see cref = "ReadProcessOutput"/> protected method).
		/// </event>
		/// <event cref = "Exited">
		/// Signals that the process has exited.
		/// </event>
		/// <remarks>
		/// <b>OnProcessExited</b> is the event handler associated with the <b>Exited</b> event
		/// raised by the <see cref="m_proc"/> process member.
		/// <para/>
		/// When a <b>CommandLineProcess</b> instance is notified that the <b>m_proc</b> process member has exited,
		/// it raises its own <b>Exited</b> event in turn, which holds the additional information of
		/// exit time and code.<br/>
		/// Before that, eventually pending text output from the process is read (by a call to 
		/// <see cref="ReadProcessOutput"/>, possibly resulting in some more <see cref="OutputLine"/>
		/// events), ensuring that no output may accidentally be lost. Then the thread 
		/// intercepting process output is stopped (via <see cref="StopOutputRecording"/>).
		/// </remarks>
		/// <seealso cref="Exited"/>
		/// <seealso cref="RaiseExitedEvent">RaiseExitedEvent (protected)</seealso>
		/// <seealso cref="m_proc">m_proc (protected)</seealso>
		/// <seealso cref="ReadProcessOutput">ReadProcessOutput (protected)</seealso>
		/// <seealso cref="StopOutputRecording">StopOutputRecording (protected)</seealso>
		/// <seealso cref="System.Diagnostics.Process.Exited">System.Diagnostics.Process.Exited</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitCode">System.Diagnostics.Process.ExitCode</seealso>
		/// <seealso cref="System.Diagnostics.Process.ExitTime">System.Diagnostics.Process.ExitTime</seealso>
		protected void OnProcessExited(object sender, EventArgs e)
		{
			int nCode		= UndefinedCodeValue;
			DateTime dtExit = UndefinedTimeValue;
			
			bool bEx = true;
			
			try
			{
				dtExit = m_proc.ExitTime;	// query process object's values
				nCode  = m_proc.ExitCode;
				bEx = false;
			}
			catch 
			{
				bEx = true;	// an error occured
			}
			finally	
			{
				ReadProcessOutput();						// Read all eventually pending text
				StopOutputRecording();						//  output before stopping.
				
				if (bEx) RaiseExitedEvent();				// Raise Exited event - either empty
					else RaiseExitedEvent(nCode, dtExit);	//  or with exit code and time.
			}
		} // OnProcessExited()


	#endregion // Implementation (private, protected, and internal methods)

	} // CommandLineProcess class

} // namespace twCompsclass CommandLineProcessDemo_form



